#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.contrib.nfp.configurator.drivers.base import base_driver
from gbpservice.contrib.nfp.configurator.drivers.loadbalancer.v1.\
    haproxy import (haproxy_rest_client)
from gbpservice.contrib.nfp.configurator.lib import constants as common_const
from gbpservice.contrib.nfp.configurator.lib import data_parser
from gbpservice.contrib.nfp.configurator.lib import lb_constants
from gbpservice.nfp.core import log as nfp_logging

LOG = nfp_logging.getLogger(__name__)

DRIVER_NAME = 'loadbalancer'


class LbGenericConfigDriver(base_driver.BaseDriver):
    """ Loadbalancer generic configuration driver class for handling device
        configuration requests.
    """

    def __init__(self):
        self.parse = data_parser.DataParser()

    def configure_interfaces(self, context, resource_data):
        """ Configure interfaces for the service VM.

        Internally it configures log forwarding in service vm
        :param context: neutron context
        :param resource_data: resource data containing service vm
                              related details

        Returns: SUCCESS/FAILED with reason.

        """

        resource_data = self.parse.parse_data(common_const.INTERFACES,
                                              resource_data)
        mgmt_ip = resource_data['mgmt_ip']

        try:
            result_log_forward = self._configure_log_forwarding(
                lb_constants.REQUEST_URL, mgmt_ip,
                self.port)
        except Exception as err:
            msg = ("Failed to configure log forwarding for service at %s. "
                   "Error: %s" % (mgmt_ip, err))
            LOG.error(msg)
        else:
            if result_log_forward == common_const.UNHANDLED:
                pass
            elif result_log_forward != lb_constants.STATUS_SUCCESS:
                msg = ("Failed to configure log forwarding for service at %s. "
                       "Error: %s" % (mgmt_ip, err))
                LOG.error(msg)
                # Failure in log forward configuration won't break chain
                # creation. However, error will be logged for detecting
                # failure.
            else:
                msg = ("Configured log forwarding for service at %s."
                       % (mgmt_ip))
                LOG.info(msg)

        return lb_constants.STATUS_SUCCESS


@base_driver.set_class_attr(SERVICE_TYPE=lb_constants.SERVICE_TYPE,
                            SERVICE_VENDOR=common_const.HAPROXY)
class HaproxyOnVmDriver(LbGenericConfigDriver):
    """Main driver which gets registered with LB agent and Generic Config agent
       in configurator and these agents pass all *aaS neutron and generic
       config requests to this class.
    """
    pool_to_device = {}

    def __init__(self, plugin_rpc=None, conf=None):
        self.plugin_rpc = plugin_rpc
        self.conf = conf
        self.timeout = 60
        self.port = lb_constants.HAPROXY_AGENT_LISTEN_PORT
        super(HaproxyOnVmDriver, self).__init__()

    def _get_rest_client(self, ip_addr):
        client = haproxy_rest_client.HttpRequests(
            ip_addr, self.port,
            lb_constants.REQUEST_RETRIES,
            lb_constants.REQUEST_TIMEOUT)
        return client

    def _get_device_for_pool(self, pool_id, context):
        resource_data = self.parse.parse_data(common_const.LOADBALANCER,
                                              context)
        role = resource_data.get('role', '')
        key = pool_id + role
        device = HaproxyOnVmDriver.pool_to_device.get(key, None)
        if device is not None:
            return device

        logical_device = self.plugin_rpc.get_logical_device(pool_id,
                                                            context)
        vip = logical_device.get('vip', None)
        if vip is None:
            return None
        else:
            device = resource_data['mgmt_ip']
            if device:
                HaproxyOnVmDriver.pool_to_device[key] = device
                return device

    def _expand_expected_codes(self, codes):
        """Expand the expected code string in set of codes.

        200-204 -> 200, 201, 202, 204
        200, 203 -> 200, 203
        """

        retval = set()
        for code in codes.replace(',', ' ').split(' '):
            code = code.strip()
            if not code:
                continue
            elif '-' in code:
                low, hi = code.split('-')[:2]
                retval.update(str(i) for i in xrange(int(low), int(hi) + 1))
            else:
                retval.add(code)
        return retval

    def _prepare_haproxy_frontend(self, vip, resource_data):
        vip_ip = vip['address']
        vip_port_number = vip['protocol_port']
        protocol = vip['protocol']

        frontend = {
            'option': {},
            'bind': '%s:%d' % (vip_ip, vip_port_number),
            'mode': lb_constants.PROTOCOL_MAP[protocol],
            'default_backend': "bck:%s" % vip['pool_id']
        }
        if vip['connection_limit'] >= 0:
            frontend.update({'maxconn': '%s' % vip['connection_limit']})
        if protocol in [lb_constants.PROTOCOL_HTTP,
                        lb_constants.PROTOCOL_HTTPS]:
            frontend['option'].update({'httplog': True})
        else:
            frontend['option'].update({'tcplog': True})
        try:
            if protocol == lb_constants.PROTOCOL_HTTP:
                frontend['option'].update({'forwardfor': True})
            provider_interface_mac = resource_data['provider_mac']
            frontend.update({'provider_interface_mac': provider_interface_mac})
        except Exception as e:
            raise e
        return frontend

    def _prepare_haproxy_backend(self, pool, context):
        logical_device = self.plugin_rpc.get_logical_device(pool['id'],
                                                            context)
        protocol = pool['protocol']
        lb_method = pool['lb_method']
        monitor = None
        for monitor in logical_device['healthmonitors']:
            break
        server_addon = ''

        backend = {
            'mode': '%s' % lb_constants.PROTOCOL_MAP[protocol],
            'balance': '%s' % lb_constants.BALANCE_MAP.get(
                lb_method, 'roundrobin'),
            'option': {},
            'timeout': {},
            'server': {}
        }
        try:
            if protocol == lb_constants.PROTOCOL_HTTP:
                backend['option'].update({'forwardfor': True})

            # health monitor options
            if monitor:
                # server addon options
                server_addon = ('check inter %(delay)ds fall %(max_retries)d'
                                % monitor)

                backend['timeout'].update({'check': '%ds'
                                           % monitor['timeout']})
                if monitor['type'] in (lb_constants.HEALTH_MONITOR_HTTP,
                                       lb_constants.HEALTH_MONITOR_HTTPS):
                    backend['option'].update(
                        {'httpchk': '%(http_method)s %(url_path)s' % monitor})
                    backend.update({'http-check expect': 'rstatus %s'
                                    % '|'.join(
                                        self._expand_expected_codes(
                                            monitor['expected_codes']))})
                if monitor['type'] == lb_constants.HEALTH_MONITOR_HTTPS:
                    backend['option'].update({'ssl-hello-chk': True})

            # session persistance options
            vip = logical_device['vip']
            persistence = vip.get('session_persistence')
            if persistence:
                if persistence['type'] == 'SOURCE_IP':
                    backend.update({'stick-table type': 'ip size 10k'})
                    backend.update({'stick on': 'src'})
                elif persistence['type'] == 'HTTP_COOKIE':
                    backend.update({'cookie': 'SRV insert indirect nocache'})
                elif (persistence['type'] == 'APP_COOKIE' and
                      persistence.get('cookie_name')):
                    backend.update({'appsession': '%s len 56 timeout 3h' %
                                    persistence['cookie_name']})

            # server options
            for member in logical_device['members']:
                backend['server'].update(
                    {"srvr:%s" % member['id']: [
                        '%(address)s:%(protocol_port)s' % member,
                        'weight %(weight)s' % member, server_addon]}
                )
                if (vip.get('session_persistence') and
                        vip['session_persistence']['type'] == 'HTTP_COOKIE'):
                    backend['server'][member['id']].append(
                        'cookie %d'
                        % logical_device['members'].index(
                            member['id']))

            return backend
        except Exception as e:
            raise e

    def _prepare_haproxy_backend_with_member(self, member, backend, context):
        logical_device = self.plugin_rpc.get_logical_device(member['pool_id'],
                                                            context)
        vip = logical_device['vip']
        monitor = None
        # chose first monitor
        for monitor in logical_device['healthmonitors']:
            break

        # update backend with the new server
        if monitor:
            server_addon = ('check inter %(delay)ds fall %(max_retries)d'
                            % monitor)
        else:
            server_addon = ''
        try:
            backend['server'].update(
                {'srvr:%s' % member['id']: [
                    '%(address)s:%(protocol_port)s' % member,
                    'weight %(weight)s' % member, server_addon]})
        except Exception as e:
            raise e
        if (vip.get('session_persistence') and
                vip['session_persistence']['type'] == 'HTTP_COOKIE'):
            backend['server'][member['id']].append(
                'cookie %d' % logical_device['members'].index(member['id']))

        return backend

    def _prepare_backend_adding_health_monitor_to_pool(self, health_monitor,
                                                       pool_id,
                                                       backend):
        # server addon options
        server_addon = ('check inter %(delay)ds fall %(max_retries)d'
                        % health_monitor)
        for server in backend['server'].itervalues():
            total_lines = len(server)
            for index, line in enumerate(server):
                if 'check' in line:
                    server[index] = server_addon
                    break
                elif total_lines == index + 1:
                    server.append(server_addon)

        try:
            backend['timeout'].update({'check': '%ds'
                                       % health_monitor['timeout']})
            if health_monitor['type'] in (lb_constants.HEALTH_MONITOR_HTTP,
                                          lb_constants.HEALTH_MONITOR_HTTPS):
                backend['option'].update(
                    {'httpchk': ('%(http_method)s %(url_path)s'
                                 % health_monitor)})
                backend.update({'http-check expect': 'rstatus %s' % (
                                '|'.join(self._expand_expected_codes(
                                    health_monitor['expected_codes'])))})
            if health_monitor['type'] == lb_constants.PROTOCOL_HTTPS:
                backend['option'].update({'ssl-hello-chk': True})
        except Exception as e:
            raise e
        return backend

    def _prepare_backend_deleting_health_monitor_from_pool(self,
                                                           health_monitor,
                                                           pool_id,
                                                           backend, context):
        logical_device = self.plugin_rpc.get_logical_device(pool_id, context)
        remaining_hms_type = []
        for monitor in logical_device['healthmonitors']:
            if health_monitor['type'] != monitor['type']:
                remaining_hms_type.append(monitor['type'])

        # Remove http, https corresponding configuration
        # Not removing http or https configuration if any 1 of them,
        # present in remaining hms type.
        try:
            if ((lb_constants.HEALTH_MONITOR_HTTP and
                    lb_constants.HEALTH_MONITOR_HTTPS)
                not in remaining_hms_type and health_monitor['type'] in
                    (lb_constants.HEALTH_MONITOR_HTTP,
                     lb_constants.HEALTH_MONITOR_HTTPS)):
                del backend['option']['httpchk']
                del backend['http-check expect']
                if health_monitor['type'] == lb_constants.HEALTH_MONITOR_HTTPS:
                    del backend['option']['ssl-hello-chk']

            server_addon = ('check inter %(delay)ds fall %(max_retries)d'
                            % health_monitor)
            for server in backend['server'].itervalues():
                for index, line in enumerate(server):
                    if 'check' in line:
                        if len(logical_device['healthmonitors']) == 0:
                            del server[index]
                        else:
                            server[index] = server_addon
                        break

            if len(logical_device['healthmonitors']) == 0:
                del backend['timeout']['check']
        except Exception as e:
            raise e
        return backend

    def _prepare_backend_updating_health_monitor_for_pool(self, health_monitor,
                                                          pool_id,
                                                          backend):
        # update backend by updatinig the health monitor
        # server addon options
        server_addon = ('check inter %(delay)ds fall %(max_retries)d'
                        % health_monitor)
        for server in backend['server'].itervalues():
            health_chk_index_in_srvr_list = 0
            for line in server:
                if 'check' in line:
                    server[health_chk_index_in_srvr_list] = server_addon
                    break
                else:
                    health_chk_index_in_srvr_list += 1

        try:
            backend['timeout'].update({'check': '%ds'
                                       % health_monitor['timeout']})
            if health_monitor['type'] in (lb_constants.HEALTH_MONITOR_HTTP,
                                          lb_constants.HEALTH_MONITOR_HTTPS):
                backend['option'].update(
                    {'httpchk': ('%(http_method)s %(url_path)s'
                                 % health_monitor)})
                backend.update({'http-check expect': 'rstatus %s' % '|'.join(
                    self._expand_expected_codes(
                        health_monitor['expected_codes']))})
            if health_monitor['type'] == lb_constants.HEALTH_MONITOR_HTTPS:
                backend['option'].update({'ssl-hello-chk': True})
        except Exception as e:
            raise e

        return backend

    def _create_vip(self, vip, device_addr, resource_data):
        try:
            client = self._get_rest_client(device_addr)
            frontend = self._prepare_haproxy_frontend(vip, resource_data)
            body = {"frnt:%s" % vip['id']: frontend}
            client.create_resource("frontend", body)
        except Exception as e:
            raise e

    def _delete_vip(self, vip, device_addr):
        try:
            client = self._get_rest_client(device_addr)
            client.delete_resource("frontend/frnt:%s" % vip['id'])
        except Exception as e:
            raise e

    def _create_pool(self, pool, device_addr, context):
        try:
            client = self._get_rest_client(device_addr)
            backend = self._prepare_haproxy_backend(pool, context)
            body = {'bck:%s' % pool['id']: backend}
            client.create_resource("backend", body)
        except Exception as e:
            raise e

    def _delete_pool(self, pool, device_addr):
        try:
            client = self._get_rest_client(device_addr)
            client.delete_resource("backend/bck:%s" % pool['id'])
        except Exception as e:
            raise e

    def _create_member(self, member, device_addr, context):
        try:
            client = self._get_rest_client(device_addr)
            backend = client.get_resource("backend/bck:%s"
                                          % member['pool_id'])
            backend = self._prepare_haproxy_backend_with_member(
                member, backend, context)
            client.update_resource("backend/bck:%s" % member['pool_id'],
                                   backend)
        except Exception as e:
            raise e

    def _delete_member(self, member, device_addr):
        try:
            client = self._get_rest_client(device_addr)
            backend = client.get_resource("backend/bck:%s"
                                          % member['pool_id'])

            # update backend with the server deleted from that
            del backend['server']['srvr:%s' % member['id']]
            client.update_resource("backend/bck:%s" % member['pool_id'],
                                   backend)
        except Exception as e:
            raise e

    def _create_pool_health_monitor(self, hm, pool_id, device_addr):
        try:
            client = self._get_rest_client(device_addr)
            backend = client.get_resource("backend/bck:%s" % pool_id)
            backend = self._prepare_backend_adding_health_monitor_to_pool(
                hm,
                pool_id,
                backend)
            client.update_resource("backend/bck:%s" % pool_id, backend)
        except Exception as e:
            raise e

    def _delete_pool_health_monitor(self, hm, pool_id,
                                    device_addr, context):
        try:
            client = self._get_rest_client(device_addr)
            backend = client.get_resource("backend/bck:%s" % pool_id)
            backend = self._prepare_backend_deleting_health_monitor_from_pool(
                hm,
                pool_id,
                backend,
                context)
            client.update_resource("backend/bck:%s" % pool_id, backend)
        except Exception as e:
            raise e

    @classmethod
    def get_name(self):
        return DRIVER_NAME

    def get_stats(self, pool_id):
        stats = {}
        try:
            # if pool is not known, do nothing
            device = HaproxyOnVmDriver.pool_to_device.get(pool_id, None)
            if device is None:
                return stats

            device_addr = self._get_device_for_pool(pool_id, None)

            # create REST client object

            client = self._get_rest_client(device_addr)
            stats = client.get_resource('stats/%s' % pool_id)

            for key, value in stats.get('members', {}).items():
                if key.find(":") != -1:
                    member_id = key[key.find(":") + 1:]
                    del stats['members'][key]
                    stats['members'][member_id] = value
        except Exception as e:
            msg = ("Failed to get stats. %s"
                   % str(e).capitalize())
            LOG.error(msg)
            raise e

        return stats

    def create_vip(self, vip, context):
        resource_data = self.parse.parse_data(common_const.LOADBALANCER,
                                              context)
        msg = ("Handling 'Create VIP' for VIP:%s with Pool:%s"
               "and tenant:%s"
               % (vip['id'], vip['pool_id'], vip['tenant_id']))
        LOG.info(msg)
        try:
            device_addr = self._get_device_for_pool(vip['pool_id'], context)
            logical_device = self.plugin_rpc.get_logical_device(vip['pool_id'],
                                                                context)

            self._create_pool(logical_device['pool'], device_addr, context)
            for member in logical_device['members']:
                self._create_member(member, device_addr, context)
            for hm in logical_device['healthmonitors']:
                self._create_pool_health_monitor(hm,
                                                 vip['pool_id'], device_addr)

            self._create_vip(vip, device_addr, resource_data)
        except Exception as e:
            msg = ("Failed to create vip %s. %s"
                   % (vip['id'], str(e).capitalize()))
            LOG.error(msg)
            raise e
        else:
            msg = ("Created vip %s." % vip['id'])
            LOG.info(msg)

    def update_vip(self, old_vip, vip, context):
        resource_data = self.parse.parse_data(common_const.LOADBALANCER,
                                              context)
        msg = ("Handling 'Update VIP' for VIP:%s and Old_VIP:%s" % (
            vip['id'], old_vip['id']))
        LOG.info(msg)
        try:
            device_addr = self._get_device_for_pool(old_vip['pool_id'],
                                                    context)

            # if old_vip is either not having associated to pool
            # or not created
            if (not old_vip['pool_id'] or
                    device_addr is None):
                return

            # is vip's pool changed
            if not vip['pool_id'] == old_vip['pool_id']:
                msg = (" VIP pool id changed to %s. Deleting old VIP:%s "
                       % (vip['pool_id'], old_vip['pool_id']))
                LOG.info(msg)
                # Delete the old VIP
                self._delete_vip(old_vip, device_addr)

                # Create the new VIP along with pool
                logical_device = self.plugin_rpc.get_logical_device(
                    vip['pool_id'],
                    context)
                pool = logical_device['pool']
                self._create_pool(pool, device_addr)
                self._create_vip(vip, device_addr, resource_data)
                return

            client = self._get_rest_client(device_addr)
            body = self._prepare_haproxy_frontend(vip, resource_data)
            client.update_resource("frontend/frnt:%s" % vip['id'], body)
        except Exception as e:
            msg = ("Failed to update vip %s. %s"
                   % (vip['id'], str(e).capitalize()))
            LOG.error(msg)
            raise e
        else:
            msg = ("Updated VIP:%s." % vip['id'])
            LOG.info(msg)

    def delete_vip(self, vip, context):
        msg = ("Handling 'Delete VIP' for VIP:%s" % (vip['id']))
        LOG.info(msg)
        try:
            device_addr = self._get_device_for_pool(vip['pool_id'], context)
            logical_device = self.plugin_rpc.get_logical_device(vip['pool_id'],
                                                                context)
            self._delete_vip(vip, device_addr)
            pool = logical_device['pool']
            self._delete_pool(pool, device_addr)
        except Exception as e:
            msg = ("Failed to delete vip %s. %s"
                   % (vip['id'], str(e).capitalize()))
            LOG.error(msg)
            raise e
        else:
            msg = ("Deleted vip %s." % vip['id'])
            LOG.info(msg)

    def create_pool(self, pool, context):
        # nothing to do here because a pool needs a vip to be useful
        msg = ("Handled 'Create Pool' for Pool:%s" % (pool['id']))
        LOG.info(msg)

    def update_pool(self, old_pool, pool, context):
        msg = ("Handling 'Update Pool' for Pool:%s and Old_Pool:%s"
               % (pool['id'], old_pool['id']))
        LOG.info(msg)
        try:
            device_addr = self._get_device_for_pool(pool['id'], context)
            if (pool['vip_id'] and
                    device_addr is not None):
                client = self._get_rest_client(device_addr)
                backend = self._prepare_haproxy_backend(pool, context)
                body = backend
                client.update_resource("backend/bck:%s" % pool['id'], body)
        except Exception as e:
            msg = ("Failed to update pool from %s to %s. %s"
                   % (old_pool['id'], pool['id'], str(e).capitalize()))
            LOG.error(msg)
            raise e
        else:
            msg = ("Updated pool from %s to %s."
                   % (old_pool['id'], pool['id']))
            LOG.info(msg)

    def delete_pool(self, pool, context):
        msg = ("Handling 'Delete Pool' for Pool:%s" % (pool['id']))
        LOG.info(msg)
        try:
            device_addr = self._get_device_for_pool(pool['id'], context)
            if device_addr is None:
                return
            if (pool['vip_id'] and
                    device_addr):
                self._delete_pool(pool, device_addr)
        except Exception as e:
            msg = ("Failed to delete pool: %s. %s"
                   % (pool['id'], str(e).capitalize()))
            LOG.error(msg)
            raise e
        else:
            msg = ("Deleted pool:%s." % pool['id'])
            LOG.info(msg)

    def create_member(self, member, context):
        msg = ("Handling 'Create Member' for Member:%s with Pool:%s "
               % (member['id'], member['pool_id']))
        LOG.info(msg)
        try:
            device_addr = self._get_device_for_pool(member['pool_id'],
                                                    context)
            if device_addr is not None:
                self._create_member(member, device_addr, context)
        except Exception as e:
            msg = ("Failed to create member %s. %s"
                   % (member['id'], str(e).capitalize()))
            LOG.error(msg)
            raise e
        else:
            msg = ("Created member %s." % member['id'])
            LOG.info(msg)

    def update_member(self, old_member, member, context):
        msg = ("Handling 'Update Member' for Member:%s with Old_Member:%s"
               % (member['id'], old_member['id']))
        LOG.info(msg)
        try:
            device_addr = self._get_device_for_pool(old_member['pool_id'],
                                                    context)
            if device_addr is not None:
                self._delete_member(old_member, device_addr)

            device_addr = self._get_device_for_pool(member['pool_id'],
                                                    context)
            if device_addr is not None:
                self._create_member(member, device_addr, context)
        except Exception as e:
            msg = ("Failed to update member %s. %s"
                   % (member['id'], str(e).capitalize()))
            LOG.error(msg)
            raise e
        else:
            msg = ("updated member %s." % member['id'])
            LOG.info(msg)

    def delete_member(self, member, context):
        msg = ("Handling 'Delete Member' for Member:%s " % (member['id']))
        LOG.info(msg)
        try:
            device_addr = self._get_device_for_pool(member['pool_id'],
                                                    context)
            if device_addr is not None:
                self._delete_member(member, device_addr)
        except Exception as e:
            msg = ("Failed to delete member %s. %s"
                   % (member['id'], str(e).capitalize()))
            LOG.error(msg)
            raise e
        else:
            msg = ("Deleted member %s." % member['id'])
            LOG.info(msg)

    def create_pool_health_monitor(self, health_monitor, pool_id, context):
        msg = ("Handling 'Create Pool Health Monitor' for "
               "Healthmonitor:%s and Pool:%s"
               % (health_monitor['id'], pool_id))
        LOG.info(msg)
        try:
            device_addr = self._get_device_for_pool(pool_id, context)
            if device_addr is not None:
                self._create_pool_health_monitor(health_monitor, pool_id,
                                                 device_addr)
        except Exception as e:
            msg = ("Failed to create pool health monitor: %s with "
                   "pool ID: %s. %s"
                   % (str(health_monitor), pool_id, str(e).capitalize()))
            LOG.error(msg)
            raise e
        else:
            msg = ("Created pool health monitor:%s with Pool: %s"
                   % (health_monitor['id'], pool_id))
            LOG.info(msg)

    def update_pool_health_monitor(self, old_health_monitor, health_monitor,
                                   pool_id, context):
        msg = ("Handling 'Update Pool Health Monitor' for HM:%s "
               "with Old_HM:%s and Pool:%s"
               % (health_monitor['id'], old_health_monitor['id'], pool_id))
        LOG.info(msg)
        try:
            device_addr = self._get_device_for_pool(pool_id, context)
            if device_addr is not None:
                client = self._get_rest_client(device_addr)
                backend = client.get_resource("backend/bck:%s" % pool_id)

                # update backend deleting the health monitor from it
                # server addon options
                backend = (
                    self._prepare_backend_updating_health_monitor_for_pool(
                        health_monitor,
                        pool_id,
                        backend))

                client.update_resource("backend/bck:%s" % pool_id, backend)
        except Exception as e:
            msg = ("Failed to update health monitor from %s to "
                   "%s for pool: %s. %s"
                   % (str(old_health_monitor), str(health_monitor),
                      pool_id, str(e).capitalize()))
            LOG.error(msg)
            raise e
        else:
            msg = ("Updated health monitor from %s to %s for Pool:%s"
                   % (old_health_monitor['id'],
                      health_monitor['id'], pool_id))
            LOG.info(msg)

    def delete_pool_health_monitor(self, health_monitor, pool_id, context):
        msg = ("Handling 'Delete Pool Health Monitor' for HM:%s Pool:%s"
               % (health_monitor['id'], pool_id))
        LOG.info(msg)
        try:
            device_addr = self._get_device_for_pool(pool_id, context)
            if device_addr is not None:
                self._delete_pool_health_monitor(health_monitor, pool_id,
                                                 device_addr, context)
        except Exception as e:
            msg = ("Failed to delete pool health monitor: %s with "
                   "pool ID: %s. %s"
                   % (str(health_monitor), pool_id, str(e).capitalize()))
            LOG.error(msg)
            raise e
        else:
            msg = ("Deleted pool health monitor: %s for Pool:%s"
                   % (health_monitor['id'], pool_id))
            LOG.info(msg)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import helpers as log

from gbpservice.neutron.services.servicechain.plugins.ncp import plumber_base


class NoopPlumber(plumber_base.NodePlumberBase):

    initialized = False

    @log.log_method_call
    def initialize(self):
        self.initialized = True

    @log.log_method_call
    def plug_services(self, context, deployment):
        self._sort_deployment(deployment)

    @log.log_method_call
    def unplug_services(self, context, deployment):
        self._sort_deployment(deployment)



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db import model_base
import sqlalchemy as sa

from gbpservice.neutron.extensions import cisco_apic
from gbpservice.neutron.extensions import cisco_apic_l3


class NetworkExtensionDb(model_base.BASEV2):

    __tablename__ = 'apic_aim_network_extensions'

    network_id = sa.Column(
        sa.String(36), sa.ForeignKey('networks.id', ondelete="CASCADE"),
        primary_key=True)
    external_network_dn = sa.Column(sa.String(1024))
    nat_type = sa.Column(sa.Enum('distributed', 'edge', ''))


class NetworkExtensionCidrDb(model_base.BASEV2):

    __tablename__ = 'apic_aim_network_external_cidrs'

    network_id = sa.Column(
        sa.String(36), sa.ForeignKey('networks.id', ondelete="CASCADE"),
        primary_key=True)
    cidr = sa.Column(sa.String(64), primary_key=True)


class SubnetExtensionDb(model_base.BASEV2):

    __tablename__ = 'apic_aim_subnet_extensions'

    subnet_id = sa.Column(
        sa.String(36), sa.ForeignKey('subnets.id', ondelete="CASCADE"),
        primary_key=True)
    snat_host_pool = sa.Column(sa.Boolean)


class RouterExtensionContractDb(model_base.BASEV2):

    __tablename__ = 'apic_aim_router_external_contracts'

    router_id = sa.Column(
        sa.String(36), sa.ForeignKey('routers.id', ondelete="CASCADE"),
        primary_key=True)
    contract_name = sa.Column(sa.String(64), primary_key=True)
    provides = sa.Column(sa.Boolean, primary_key=True)


class ExtensionDbMixin(object):

    def _set_if_not_none(self, res_dict, res_attr, db_attr):
        if db_attr is not None:
            res_dict[res_attr] = db_attr

    def get_network_extn_db(self, session, network_id):
        db_obj = (session.query(NetworkExtensionDb).filter_by(
                  network_id=network_id).first())
        db_cidrs = (session.query(NetworkExtensionCidrDb).filter_by(
                    network_id=network_id).all())
        result = {}
        if db_obj:
            self._set_if_not_none(result, cisco_apic.EXTERNAL_NETWORK,
                                  db_obj['external_network_dn'])
            self._set_if_not_none(result, cisco_apic.NAT_TYPE,
                                  db_obj['nat_type'])
        if result.get(cisco_apic.EXTERNAL_NETWORK):
            result[cisco_apic.EXTERNAL_CIDRS] = [c['cidr'] for c in db_cidrs]
        return result

    def set_network_extn_db(self, session, network_id, res_dict):
        with session.begin(subtransactions=True):
            db_obj = (session.query(NetworkExtensionDb).filter_by(
                      network_id=network_id).first())
            db_obj = db_obj or NetworkExtensionDb(network_id=network_id)
            if cisco_apic.EXTERNAL_NETWORK in res_dict:
                db_obj['external_network_dn'] = (
                    res_dict[cisco_apic.EXTERNAL_NETWORK])
            if cisco_apic.NAT_TYPE in res_dict:
                db_obj['nat_type'] = res_dict[cisco_apic.NAT_TYPE]
            session.add(db_obj)
            if cisco_apic.EXTERNAL_CIDRS in res_dict:
                self._update_list_attr(session, NetworkExtensionCidrDb, 'cidr',
                                       res_dict[cisco_apic.EXTERNAL_CIDRS],
                                       network_id=network_id)

    def get_subnet_extn_db(self, session, subnet_id):
        db_obj = (session.query(SubnetExtensionDb).filter_by(
                  subnet_id=subnet_id).first())
        result = {}
        if db_obj:
            self._set_if_not_none(result, cisco_apic.SNAT_HOST_POOL,
                                  db_obj['snat_host_pool'])
        return result

    def set_subnet_extn_db(self, session, subnet_id, res_dict):
        db_obj = (session.query(SubnetExtensionDb).filter_by(
                  subnet_id=subnet_id).first())
        db_obj = db_obj or SubnetExtensionDb(subnet_id=subnet_id)
        if cisco_apic.SNAT_HOST_POOL in res_dict:
            db_obj['snat_host_pool'] = res_dict[cisco_apic.SNAT_HOST_POOL]
        session.add(db_obj)

    def get_router_extn_db(self, session, router_id):
        db_contracts = (session.query(RouterExtensionContractDb).filter_by(
                        router_id=router_id).all())
        return {cisco_apic_l3.EXTERNAL_PROVIDED_CONTRACTS:
                [c['contract_name'] for c in db_contracts if c['provides']],
                cisco_apic_l3.EXTERNAL_CONSUMED_CONTRACTS:
                [c['contract_name'] for c in db_contracts
                 if not c['provides']]}

    def _update_list_attr(self, session, db_model, column,
                          new_values, **filters):
        rows = session.query(db_model).filter_by(**filters).all()
        new_values = set(new_values)
        for r in rows:
            if r[column] in new_values:
                new_values.discard(r[column])
            else:
                session.delete(r)
        for v in new_values:
            attr = {column: v}
            attr.update(filters)
            db_obj = db_model(**attr)
            session.add(db_obj)

    def set_router_extn_db(self, session, router_id, res_dict):
        with session.begin(subtransactions=True):
            if cisco_apic_l3.EXTERNAL_PROVIDED_CONTRACTS in res_dict:
                self._update_list_attr(session, RouterExtensionContractDb,
                   'contract_name',
                   res_dict[cisco_apic_l3.EXTERNAL_PROVIDED_CONTRACTS],
                   router_id=router_id, provides=True)
            if cisco_apic_l3.EXTERNAL_CONSUMED_CONTRACTS in res_dict:
                self._update_list_attr(session, RouterExtensionContractDb,
                    'contract_name',
                    res_dict[cisco_apic_l3.EXTERNAL_CONSUMED_CONTRACTS],
                    router_id=router_id, provides=False)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.neutron.services.grouppolicy.common import exceptions as gpexc


class InvalidApicName(gpexc.GroupPolicyBadRequest):
    message = _("Resource has no valid APIC name.")

APIC_REFERENCE_PREFIX = 'apic:'


class ApicNameManager(object):

    gbp_to_apic = {'l3_policy': 'context',
                   'l2_policy': 'bridge_domain',
                   'policy_target_group': 'endpoint_group',
                   'policy_rule_set': 'contract'}

    def __init__(self, apic_manager):
        self.name_mapper = apic_manager.apic_mapper
        self.dn_manager = apic_manager.apic.dn_manager

    def __getattr__(self, item):
        if self.name_mapper.is_valid_name_type(item):
            def get_name_wrapper(context, obj_id, prefix=''):
                return self._get_name(item, context, obj_id, prefix=prefix)
            return get_name_wrapper

        raise AttributeError

    def tenant(self, obj):
        if self._is_apic_reference(obj):
            parts = self._try_all_types(obj)
            if parts:
                return parts[0]
        return self.name_mapper.tenant(None, obj['tenant_id'])

    def has_valid_name(self, obj):
        if self._is_apic_reference(obj):
            if not self._try_all_types(obj):
                raise InvalidApicName()

    def _try_all_types(self, obj):
        for possible in self.dn_manager.nice_to_rn:
            parts = getattr(self.dn_manager, 'decompose_%s' % possible)(
                self._extract_apic_reference(obj))
            if parts:
                return parts

    def _get_name(self, obj_type, context, obj, prefix=''):
        if self._is_apic_reference(obj) and obj_type in self.gbp_to_apic:
            map_type = self.gbp_to_apic[obj_type]
            parts = getattr(self.dn_manager, 'decompose_%s' % map_type)(
                self._extract_apic_reference(obj))
            result = self.name_mapper.pre_existing(context, parts[-1])
        else:
            obj_id = (obj['reuse_bd']
                      if obj_type == 'l2_policy' and obj.get('reuse_bd')
                      else obj['id'])
            result = getattr(self.name_mapper, obj_type)(context, obj_id,
                                                         prefix=prefix)
        return result

    def _is_apic_reference(self, obj):
        return obj['name'].startswith(APIC_REFERENCE_PREFIX)

    def _extract_apic_reference(self, obj):
        return obj['name'][len(APIC_REFERENCE_PREFIX):]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class ServiceChainContext(object):
    """ServiceChain context base class."""
    def __init__(self, plugin, plugin_context):
        self._plugin = plugin
        self._plugin_context = plugin_context


class ServiceChainNodeContext(ServiceChainContext):

    def __init__(self, plugin, plugin_context, sc_node,
                 original_sc_node=None):
        super(ServiceChainNodeContext, self).__init__(plugin, plugin_context)
        self._sc_node = sc_node
        self._profile = None
        if self._sc_node['service_profile_id']:
            self._profile = self._plugin.get_service_profile(
                self._plugin_context, self._sc_node['service_profile_id'])

        self._original_sc_node = original_sc_node
        self._original_profile = None
        if (self._original_sc_node and
                self._original_sc_node['service_profile_id']):
            self._original_profile = self._plugin.get_service_profile(
                self._plugin_context,
                self._original_sc_node['service_profile_id'])

    @property
    def current(self):
        return self._sc_node

    @property
    def original(self):
        return self._original_sc_node

    @property
    def current_profile(self):
        return self._profile

    @property
    def original_profile(self):
        return self._original_profile


class ServiceChainSpecContext(ServiceChainContext):

    def __init__(self, plugin, plugin_context, sc_spec,
                 original_sc_spec=None):
        super(ServiceChainSpecContext, self).__init__(plugin, plugin_context)
        self._sc_spec = sc_spec
        self._original_sc_spec = original_sc_spec

    @property
    def current(self):
        return self._sc_spec

    @property
    def original(self):
        return self._original_sc_spec


class ServiceChainInstanceContext(ServiceChainContext):

    def __init__(self, plugin, plugin_context, sc_instance,
                 original_sc_instance=None):
        super(ServiceChainInstanceContext, self).__init__(plugin,
                                                          plugin_context)
        self._sc_instance = sc_instance
        self._original_sc_instance = original_sc_instance

    @property
    def current(self):
        return self._sc_instance

    @property
    def original(self):
        return self._original_sc_instance


class ServiceProfileContext(ServiceChainContext):

    def __init__(self, plugin, plugin_context, profile,
                 original_profile=None):
        super(ServiceProfileContext, self).__init__(plugin, plugin_context)
        self._profile = profile
        self._original_profile = original_profile

    @property
    def current(self):
        return self._profile

    @property
    def original(self):
        return self._original_profile



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""segmentation labels for PT (apic)

Revision ID: 092e4b1aeb0a
Revises: d4bb487a81b8
Create Date: 2016-09-18 18:58:26.810742

"""

# revision identifiers, used by Alembic.
revision = '092e4b1aeb0a'
down_revision = 'd4bb487a81b8'

from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'gp_apic_mapping_segmentation_labels',
        sa.Column('policy_target_id', sa.String(length=36), nullable=False),
        sa.Column('segmentation_label', sa.String(length=255),
                  nullable=False),
        sa.ForeignKeyConstraint(
            ['policy_target_id'], ['gp_policy_targets.id'],
            name='gp_apic_mapping_segmentation_lablel_fk_ptid',
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('policy_target_id', 'segmentation_label')
    )


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Reuse BD

Revision ID: 75aa8a37a8de
Revises: a707faecf518
Create Date: 2016-11-11 17:01:30.735865

"""

# revision identifiers, used by Alembic.
revision = '75aa8a37a8de'
down_revision = 'a707faecf518'

from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'gp_apic_mapping_reuse_bds',
        sa.Column('l2_policy_id', sa.String(length=36), nullable=False),
        sa.Column('target_l2_policy_id', sa.String(length=36),
                  nullable=False),
        sa.ForeignKeyConstraint(
            ['l2_policy_id'], ['gp_l2_policies.id'],
            name='gp_apic_mapping_reuse_bd_fk_l2pid',
            ondelete='CASCADE'),
        sa.ForeignKeyConstraint(
            ['target_l2_policy_id'], ['gp_l2_policies.id'],
            name='gp_apic_mapping_reuse_bd_target_l2p_id_fk_l2pid'),
        sa.PrimaryKeyConstraint('l2_policy_id')
    )


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from neutron.common import constants as n_constants

from gbpservice.neutron.db.grouppolicy import group_policy_mapping_db as gpdb
from gbpservice.neutron.services.grouppolicy.common import constants as g_const
from gbpservice.neutron.services.grouppolicy.common import exceptions as gpexc


ALLOWING_ACTIONS = [g_const.GP_ACTION_ALLOW, g_const.GP_ACTION_REDIRECT]
REVERSE_PREFIX = 'reverse-'
SERVICE_PREFIX = 'Svc-'
IMPLICIT_PREFIX = 'implicit-'
PER_PROJECT = 'per-project'
REVERSIBLE_PROTOCOLS = [n_constants.PROTO_NAME_TCP.lower(),
                        n_constants.PROTO_NAME_UDP.lower(),
                        n_constants.PROTO_NAME_ICMP.lower()]
ICMP_REPLY_TYPES = ['echo-rep', 'dst-unreach', 'src-quench', 'time-exceeded']
CP_ENTRY = 'os-entry'


class ExplicitSubnetAssociationNotSupported(gpexc.GroupPolicyBadRequest):
    message = _("Explicit subnet association not supported by APIC driver.")


class HierarchicalContractsNotSupported(gpexc.GroupPolicyBadRequest):
    message = _("Hierarchical contracts not supported by APIC driver.")


class MultipleExternalPoliciesForL3Policy(gpexc.GroupPolicyBadRequest):
    message = _("Potential association of multiple external policies to "
                "an L3 Policy.")


class SharedExternalPolicyUnsupported(gpexc.GroupPolicyBadRequest):
    message = _("APIC mapping driver does not support sharing of "
                "external policies.")


class OnlyOneL3PolicyIsAllowedPerExternalSegment(gpexc.GroupPolicyBadRequest):
    message = _("Only one L3 Policy per ES is supported when NAT is disabled "
                "on the ES.")


class OnlyOneAddressIsAllowedPerExternalSegment(gpexc.GroupPolicyBadRequest):
    message = _("Only one ip address on each ES is supported on "
                "APIC GBP driver.")


def get_filter_entries_for_policy_classifier(classifier):
    # forward_rules and reverse_rules is each a dict of filter_entries
    # with each entry in the dict having the filter_entry name as the
    # key and the filter_entry attributes as the value
    entries = {'forward_rules': None, 'reverse_rules': None}
    x = 0
    port_min, port_max = (
        gpdb.GroupPolicyMappingDbPlugin._get_min_max_ports_from_range(
            classifier['port_range']))
    f_attrs = {'etherT': 'unspecified'}
    if classifier['protocol']:
        f_attrs['etherT'] = 'ip'
        f_attrs['prot'] = classifier['protocol'].lower()
    if port_min and port_max:
        f_attrs['dToPort'] = port_max
        f_attrs['dFromPort'] = port_min
    entries['forward_rules'] = {_get_filter_entry_name(x): f_attrs}
    # Also create reverse rule
    if f_attrs.get('prot') in REVERSIBLE_PROTOCOLS:
        r_entries = {}
        r_attrs = copy.deepcopy(f_attrs)
        if r_attrs.get('dToPort') and r_attrs.get('dFromPort'):
            r_attrs.pop('dToPort')
            r_attrs.pop('dFromPort')
            r_attrs['sToPort'] = port_max
            r_attrs['sFromPort'] = port_min
        if r_attrs['prot'] == n_constants.PROTO_NAME_TCP.lower():
            # Only match on established sessions
            r_attrs['tcpRules'] = 'est'
        r_entries[_get_filter_entry_name(x)] = r_attrs
        if r_attrs['prot'] == n_constants.PROTO_NAME_ICMP.lower():
            r_entries = {}
            # create more entries:
            for reply_type in ICMP_REPLY_TYPES:
                x += 1
                r_entry = copy.deepcopy(r_attrs)
                r_entry['icmpv4T'] = reply_type
                r_entries[_get_filter_entry_name(x)] = r_entry
        entries['reverse_rules'] = r_entries
    return entries


def get_filter_entries_for_policy_rule(context):
    # forward_rules and reverse_rules is each a dict of filter_entries
    # with each entry in the dict having the filter_entry name as the
    # key and the filter_entry attributes as the value
    entries = {'forward_rules': None, 'reverse_rules': None}
    action = context._plugin.get_policy_action(
        context._plugin_context, context.current['policy_actions'][0])
    classifier = context._plugin.get_policy_classifier(
        context._plugin_context,
        context.current['policy_classifier_id'])
    if action['action_type'] in ALLOWING_ACTIONS:
        entries = get_filter_entries_for_policy_classifier(classifier)
    return entries


def get_arp_filter_entry():
    return {'arp': {'etherT': 'arp'}}


def get_service_contract_filter_entries():
    entries = {}
    # DNS
    dns_attrs = {'etherT': 'ip',
                 'prot': 'udp',
                 'dToPort': 'dns',
                 'dFromPort': 'dns'}
    entries['dns'] = dns_attrs
    r_dns_attrs = {'etherT': 'ip',
                   'prot': 'udp',
                   'sToPort': 'dns',
                   'sFromPort': 'dns'}
    entries['r-dns'] = r_dns_attrs

    # HTTP
    http_attrs = {'etherT': 'ip',
                  'prot': 'tcp',
                  'dToPort': 80,
                  'dFromPort': 80}
    entries['http'] = http_attrs
    r_http_attrs = {'etherT': 'ip',
                    'prot': 'tcp',
                    'sToPort': 80,
                    'sFromPort': 80}
    entries['r-http'] = r_http_attrs

    icmp_attrs = {'etherT': 'ip',
                  'prot': 'icmp'}
    entries['icmp'] = icmp_attrs

    icmpv6_attrs = {'etherT': 'ip',
                    'prot': 58}
    entries['icmpv6'] = icmpv6_attrs

    # DHCP
    dhcp_attrs = {'etherT': 'ip',
                  'prot': 'udp',
                  'dToPort': 68,
                  'dFromPort': 68,
                  'sToPort': 67,
                  'sFromPort': 67}
    entries['dhcp'] = dhcp_attrs
    r_dhcp_attrs = {'etherT': 'ip',
                    'prot': 'udp',
                    'dToPort': 67,
                    'dFromPort': 67,
                    'sToPort': 68,
                    'sFromPort': 68}
    entries['r-dhcp'] = r_dhcp_attrs

    dhcpv6_attrs = {'etherT': 'ip',
                    'prot': 'udp',
                    'dToPort': 546,
                    'dFromPort': 546,
                    'sToPort': 547,
                    'sFromPort': 547}
    entries['dhcpv6'] = dhcpv6_attrs
    r_dhcpv6_attrs = {'etherT': 'ip',
                      'prot': 'udp',
                      'dToPort': 547,
                      'dFromPort': 547,
                      'sToPort': 546,
                      'sFromPort': 546}
    entries['r-dhcpv6'] = r_dhcpv6_attrs

    # ARP
    arp_entries = get_arp_filter_entry()
    for k, v in arp_entries.iteritems():
        entries[k] = v

    return entries


def map_to_aim_filter_entry(entry):
    mapped_keys = {'etherT': 'ether_type',
                   'prot': 'ip_protocol',
                   'dToPort': 'dest_to_port',
                   'dFromPort': 'dest_from_port',
                   'sToPort': 'source_to_port',
                   'sFromPort': 'source_from_port',
                   'icmpv4T': 'icmpv4_type',
                   'tcpRules': 'tcp_flags'}
    return dict((mapped_keys[k], v) for (k, v) in entry.iteritems())


def _get_filter_entry_name(entry_number):
    return CP_ENTRY + '-' + str(entry_number)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import heapq
import signal
from time import time

from gbpservice.nfp.core import log as nfp_logging


LOG = nfp_logging.getLogger(__name__)

alarmlist = []

__new_alarm = lambda t, f, a, k: (t + time(), f, a, k)
__next_alarm = lambda: int(
    round(alarmlist[0][0] - time())) if alarmlist else None
__set_alarm = lambda: signal.alarm(max(__next_alarm(), 1))


class Watchdog(object):

    def __init__(self, callback, seconds=20 * 60, **kwargs):
        self._seconds = seconds
        self._callback = callback
        self.kwargs = kwargs

        self._alarm = alarm(self._seconds, self.timedout)

    def timedout(self):
        try:
            self._callback(**self.kwargs)
        except Exception as e:
            message = "Unexpected exception - %s" % (e)
            LOG.error(message)

    def cancel(self):
        try:
            cancel(self._alarm)
        except ValueError:
            pass
        except Exception as e:
            message = "Unexpected exception - %s" % (e)
            LOG.error(message)


def __clear_alarm():
    """Clear an existing alarm.

    If the alarm signal was set to a callable other than our own, queue the
    previous alarm settings.
    """
    oldsec = signal.alarm(0)
    oldfunc = signal.signal(signal.SIGALRM, __alarm_handler)
    if oldsec > 0 and oldfunc != __alarm_handler:
        heapq.heappush(alarmlist, (__new_alarm(oldsec, oldfunc, [], {})))


def __alarm_handler(*zargs):
    """Handle an alarm by calling any due heap entries and resetting the alarm.

    Note that multiple heap entries might get called, especially if calling an
    entry takes a lot of time.
    """
    try:
        nextt = __next_alarm()
        while nextt is not None and nextt <= 0:
            (tm, func, args, keys) = heapq.heappop(alarmlist)
            func(*args, **keys)
            nextt = __next_alarm()
    finally:
        if alarmlist:
            __set_alarm()


def alarm(sec, func, *args, **keys):
    """Set an alarm.

    When the alarm is raised in `sec` seconds, the handler will call `func`,
    passing `args` and `keys`. Return the heap entry (which is just a big
    tuple), so that it can be cancelled by calling `cancel()`.
    """
    __clear_alarm()
    try:
        newalarm = __new_alarm(sec, func, args, keys)
        heapq.heappush(alarmlist, newalarm)
        return newalarm
    finally:
        __set_alarm()


def cancel(alarm):
    """Cancel an alarm by passing the heap entry returned by `alarm()`.

    It is an error to try to cancel an alarm which has already occurred.
    """
    __clear_alarm()
    try:
        alarmlist.remove(alarm)
        heapq.heapify(alarmlist)
    finally:
        if alarmlist:
            __set_alarm()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webob.exc

from neutron import context as nctx
from neutron.db import api as db_api
from neutron.db import model_base
from neutron.tests.unit.extensions import test_l3
from neutron.tests.unit import testlib_api

from gbpservice.neutron.db.grouppolicy import group_policy_mapping_db as gpmdb
from gbpservice.neutron.services.grouppolicy.common import exceptions as gpexc
from gbpservice.neutron.tests.unit.db.grouppolicy import (
    test_group_policy_db as tgpdb)


class GroupPolicyMappingDBTestPlugin(gpmdb.GroupPolicyMappingDbPlugin):

    supported_extension_aliases = ['group-policy', 'group-policy-mapping'] + (
        tgpdb.UNSUPPORTED_REQUIRED_EXTS)
    path_prefix = "/grouppolicy"


DB_GP_PLUGIN_KLASS = (GroupPolicyMappingDBTestPlugin.__module__ + '.' +
                      GroupPolicyMappingDBTestPlugin.__name__)

SC_PLUGIN_KLASS = (
    "gbpservice.neutron.services.servicechain.plugins.msc.plugin."
    "ServiceChainPlugin")


class GroupPolicyMappingDbTestCase(tgpdb.GroupPolicyDbTestCase,
                                   test_l3.L3NatTestCaseMixin):

    def setUp(self, core_plugin=None, l3_plugin=None, gp_plugin=None,
              service_plugins=None, sc_plugin=None):
        testlib_api.SqlTestCase._TABLES_ESTABLISHED = False
        if not gp_plugin:
            gp_plugin = DB_GP_PLUGIN_KLASS
        if not service_plugins:
            service_plugins = {
                'gp_plugin_name': gp_plugin,
                'servicechain_plugin': sc_plugin or SC_PLUGIN_KLASS}
        service_plugins['l3_plugin_name'] = l3_plugin or "router"
        super(GroupPolicyMappingDbTestCase, self).setUp(
            core_plugin=core_plugin, gp_plugin=gp_plugin,
            service_plugins=service_plugins
        )
        engine = db_api.get_engine()
        model_base.BASEV2.metadata.create_all(engine)


class TestMappedGroupResources(GroupPolicyMappingDbTestCase,
                               tgpdb.TestGroupResources):
    pass


class TestMappedGroupResourceAttrs(GroupPolicyMappingDbTestCase):

    def test_create_delete_policy_target_with_port(self):
        with self.port() as port:
            port_id = port['port']['id']
            pt = self.create_policy_target(port_id=port_id)
            pt_id = pt['policy_target']['id']
            self.assertEqual(port_id, pt['policy_target']['port_id'])
            req = self.new_show_request('policy_targets', pt_id, fmt=self.fmt)
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertEqual(port_id, res['policy_target']['port_id'])
            req = self.new_delete_request('policy_targets', pt_id)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_create_delete_policy_target_group_with_subnets(self):
        with self.subnet(cidr='10.10.1.0/24') as subnet1:
            with self.subnet(cidr='10.10.2.0/24') as subnet2:
                subnets = [subnet1['subnet']['id'], subnet2['subnet']['id']]
                ptg = self.create_policy_target_group(subnets=subnets)
                ptg_id = ptg['policy_target_group']['id']
                self.assertEqual(sorted(subnets),
                                 sorted(ptg['policy_target_group']['subnets']))
                req = self.new_show_request('policy_target_groups', ptg_id,
                                            fmt=self.fmt)
                res = self.deserialize(
                    self.fmt, req.get_response(self.ext_api))
                self.assertEqual(sorted(subnets),
                                 sorted(res['policy_target_group']['subnets']))
                req = self.new_delete_request('policy_target_groups', ptg_id)
                res = req.get_response(self.ext_api)
                self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_update_policy_target_group_subnets(self):
        with self.subnet(cidr='10.10.1.0/24') as subnet1:
            with self.subnet(cidr='10.10.2.0/24') as subnet2:
                with self.subnet(cidr='10.10.3.0/24') as subnet3:
                    orig_subnets = [subnet1['subnet']['id'],
                                    subnet2['subnet']['id']]
                    ptg = self.create_policy_target_group(subnets=orig_subnets)
                    ptg_id = ptg['policy_target_group']['id']
                    self.assertEqual(
                        sorted(orig_subnets),
                        sorted(ptg['policy_target_group']['subnets']))
                    new_subnets = [subnet1['subnet']['id'],
                                   subnet3['subnet']['id']]
                    data = {'policy_target_group': {'subnets': new_subnets}}
                    req = self.new_update_request('policy_target_groups', data,
                                                  ptg_id)
                    res = self.deserialize(self.fmt,
                                           req.get_response(self.ext_api))
                    self.assertEqual(
                        sorted(new_subnets),
                        sorted(res['policy_target_group']['subnets']))
                    req = self.new_show_request('policy_target_groups', ptg_id,
                                                fmt=self.fmt)
                    res = self.deserialize(
                        self.fmt, req.get_response(self.ext_api))
                    self.assertEqual(
                        sorted(new_subnets),
                        sorted(res['policy_target_group']['subnets']))
                    # REVISIT(rkukura): Remove delete once subnet() context
                    # manager is replaced with a function that does not delete
                    # the resource(s) that are created.
                    req = self.new_delete_request('policy_target_groups',
                                                  ptg_id)
                    res = req.get_response(self.ext_api)
                    self.assertEqual(webob.exc.HTTPNoContent.code,
                                     res.status_int)

    def test_create_delete_l2_policy_with_network(self):
        with self.network() as network:
            network_id = network['network']['id']
            l2p = self.create_l2_policy(network_id=network_id)
            l2p_id = l2p['l2_policy']['id']
            self.assertEqual(network_id, l2p['l2_policy']['network_id'])
            req = self.new_show_request('l2_policies', l2p_id, fmt=self.fmt)
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertEqual(network_id, res['l2_policy']['network_id'])
            req = self.new_delete_request('l2_policies', l2p_id)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_create_delete_l3_policy_with_routers(self):
        with self.router() as router1:
            with self.router() as router2:
                routers = [router1['router']['id'], router2['router']['id']]
                l3p = self.create_l3_policy(routers=routers)
                l3p_id = l3p['l3_policy']['id']
                test_l3p_id = self._gbp_plugin.get_l3p_id_from_router_id(
                    nctx.get_admin_context(),
                    router1['router']['id'])
                self.assertEqual(l3p_id, test_l3p_id)
                test_l3p_id = self._gbp_plugin.get_l3p_id_from_router_id(
                    nctx.get_admin_context(),
                    router2['router']['id'])
                self.assertEqual(l3p_id, test_l3p_id)
                self.assertEqual(sorted(routers),
                                 sorted(l3p['l3_policy']['routers']))
                req = self.new_show_request('l3_policies', l3p_id,
                                            fmt=self.fmt)
                res = self.deserialize(self.fmt,
                                       req.get_response(self.ext_api))
                self.assertEqual(sorted(routers),
                                 sorted(res['l3_policy']['routers']))
                req = self.new_delete_request('l3_policies', l3p_id)
                res = req.get_response(self.ext_api)
                self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_update_l3_policy_routers(self):
        with self.router() as router1:
            with self.router() as router2:
                with self.router() as router3:
                    orig_routers = [router1['router']['id'],
                                    router2['router']['id']]
                    l3p = self.create_l3_policy(routers=orig_routers)
                    l3p_id = l3p['l3_policy']['id']
                    self.assertEqual(sorted(orig_routers),
                                     sorted(l3p['l3_policy']['routers']))
                    new_routers = [router1['router']['id'],
                                   router3['router']['id']]
                    data = {'l3_policy': {'routers': new_routers}}
                    req = self.new_update_request('l3_policies', data, l3p_id)
                    res = self.deserialize(self.fmt,
                                           req.get_response(self.ext_api))
                    self.assertEqual(sorted(new_routers),
                                     sorted(res['l3_policy']['routers']))
                    req = self.new_show_request('l3_policies',
                                                l3p_id, fmt=self.fmt)
                    res = self.deserialize(
                        self.fmt, req.get_response(self.ext_api))
                    self.assertEqual(sorted(new_routers),
                                     sorted(res['l3_policy']['routers']))
                    # REVISIT(rkukura): Remove delete once router() context
                    # manager is replaced with a function that does not delete
                    # the resource(s) that are created.
                    req = self.new_delete_request('l3_policies', l3p_id)
                    res = req.get_response(self.ext_api)
                    self.assertEqual(webob.exc.HTTPNoContent.code,
                                     res.status_int)

    def test_create_delete_es_with_subnet(self):
        with self.subnet(cidr='10.10.1.0/24') as subnet:
            subnet_id = subnet['subnet']['id']
            es = self.create_external_segment(subnet_id=subnet_id,
                                              expected_res_status=201)
            self.assertEqual(subnet_id, es['external_segment']['subnet_id'])
            es_id = es['external_segment']['id']
            req = self.new_show_request('external_segments', es_id,
                                        fmt=self.fmt)
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertEqual(subnet_id, res['external_segment']['subnet_id'])
            req = self.new_delete_request('external_segments', es_id)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_list_policy_targets(self):
        with self.port() as port1:
            with self.port() as port2:
                ports = [port1['port']['id'], port2['port']['id']]
                pts = [self.create_policy_target(port_id=ports[0]),
                       self.create_policy_target(port_id=ports[1])]
                self._test_list_resources('policy_target', pts)
                self._test_list_resources('policy_target', [pts[0]],
                                          query_params='port_id=' + ports[0])

    def test_list_l2_policies(self):
        with self.network() as network1:
            with self.network() as network2:
                networks = [network1['network']['id'],
                            network2['network']['id']]
                l2_policies = [self.create_l2_policy(network_id=networks[0]),
                               self.create_l2_policy(network_id=networks[1])]
                self._test_list_resources(
                    'l2_policy', l2_policies)
                self._test_list_resources(
                    'l2_policy', [l2_policies[0]],
                    query_params='network_id=' + networks[0])

    def test_list_es(self):
        with self.subnet(cidr='10.10.1.0/24') as subnet1:
            with self.subnet(cidr='10.10.2.0/24') as subnet2:
                subnets = [subnet1['subnet']['id'], subnet2['subnet']['id']]
                external_segments = [
                            self.create_external_segment(subnet_id=subnets[0]),
                            self.create_external_segment(subnet_id=subnets[1])]
                self._test_list_resources(
                    'external_segment', external_segments)
                self._test_list_resources(
                    'external_segment', [external_segments[0]],
                    query_params='subnet_id=' + subnets[0])

    def test_pt_port_extra_attributes_fail(self):
        ptg = self.create_policy_target_group()['policy_target_group']
        ctx = nctx.get_admin_context()
        self.assertRaises(gpexc.InvalidPortExtraAttributes,
                          self._gbp_plugin.create_policy_target,
                          ctx, {'policy_target': {
                                    'description': '', 'name': '',
                                    'port_id': None, 'cluster_id': '',
                                    'policy_target_group_id': ptg['id'],
                                    'port_attributes': {'network_id': ''},
                                    'tenant_id': self._tenant_id}})

        self.assertRaises(gpexc.InvalidPortExtraAttributes,
                          self._gbp_plugin.create_policy_target,
                          ctx, {'policy_target': {
                                    'description': '', 'name': '',
                                    'port_id': None, 'cluster_id': '',
                                    'policy_target_group_id': ptg['id'],
                                    'port_attributes': {
                                        'allowed_address_pairs': ''},
                                    'tenant_id': self._tenant_id}})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet
eventlet.monkey_patch()

from gbpservice.nfp.core import log as nfp_logging
import os
from oslo_config import cfg as oslo_config
from oslo_log import log as oslo_logging
import socket
import sys
import time


oslo_logging.register_options(oslo_config.CONF)

if not hasattr(oslo_config.CONF, 'module'):
    module_opts = [
        oslo_config.StrOpt('module',
                           default='proxy',
                           help='component name for logging.')]
    oslo_config.CONF.register_opts(module_opts)

LOG = nfp_logging.getLogger(__name__)

# Queue of proxy connections which workers will handle
ConnQ = eventlet.queue.Queue(maxsize=0)

tcp_open_connection_count = 0
tcp_close_connection_count = 0


class ConnectionIdleTimeOut(Exception):

    '''
    Exception raised when connection is idle for configured timeout
    '''
    pass


"""
parsing the proxy configuration file
"""


class Configuration(object):

    def __init__(self, conf):
        self.unix_bind_path = conf.proxy.unix_bind_path
        self.max_connections = conf.proxy.max_connections
        self.worker_threads = conf.proxy.worker_threads
        self.connect_max_wait_timeout = conf.proxy.connect_max_wait_timeout
        self.idle_max_wait_timeout = conf.proxy.idle_max_wait_timeout
        self.idle_min_wait_timeout = conf.proxy.idle_min_wait_timeout
        self.rest_server_address = conf.proxy.nfp_controller_ip
        self.rest_server_port = conf.proxy.nfp_controller_port


"""
Class to create Unix Listener
"""


class UnixServer(object):

    def __init__(self, conf, proxy):
        self.proxy = proxy
        self.bind_path = conf.unix_bind_path
        self.max_connections = conf.max_connections
        # Make sure the socket does not already exist
        try:
            os.unlink(self.bind_path)
        except OSError:
            if os.path.exists(self.bind_path):
                raise

        # Create a UDS socket
        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        # Bind the socket to the port
        message = 'starting up on %s' % self.bind_path
        LOG.info(message)
        self.socket.bind(self.bind_path)
        self.socket.listen(self.max_connections)

    def listen(self):
        client, address = self.socket.accept()
        self.proxy.new_client(client, address)


"""
Class to create TCP client Connection if
TCP server is alive
"""


class TcpClient(object):

    def __init__(self, conf, proxy):
        self.conf = conf
        self.proxy = proxy
        self.server_address = conf.rest_server_address
        self.server_port = conf.rest_server_port
        # Connect the socket to the port where the server is listening
        self.server = (self.server_address, self.server_port)

    def connect(self):
        sock = socket.socket()
        message = 'connecting to %s port %s' % self.server
        LOG.info(message)
        sock.settimeout(self.conf.connect_max_wait_timeout)
        try:
            sock.connect(self.server)
        except socket.error as exc:
            message = "Caught exception socket.error : %s" % exc
            LOG.error(message)
            return sock, False
        return sock, True


"""
ADT for proxy connection
"""


class Connection(object):

    def __init__(self, conf, socket, type='unix'):
        self._socket = socket
        self._idle_wait = conf.idle_min_wait_timeout
        self._idle_timeout = conf.idle_max_wait_timeout
        self._idle_count_max = (self._idle_timeout / self._idle_wait)
        self._idle_count = 0
        self._start_time = time.time()
        self._end_time = time.time()
        self.type = type
        self.socket_id = self._socket.fileno()

    def _tick(self):
        self._idle_count += 1

    def _timedout(self):
        if self._idle_count > self._idle_count_max:
            self._end_time = time.time()
            raise ConnectionIdleTimeOut(
                "Connection (%d) - stime (%s) - etime (%s) - "
                "idle_count (%d) idle_count_max(%d)" % (
                    self.identify(), self._start_time,
                    self._end_time, self._idle_count, self._idle_count_max))

    def idle(self):
        self._tick()
        self._timedout()

    def idle_reset(self):
        self._idle_count = 0
        self._start_time = time.time()

    def _wait(self, timeout):
        if self.type == 'unix':
            eventlet.sleep(timeout)
            self._socket.setblocking(0)
        else:
            self._socket.settimeout(timeout)

    def recv(self):
        self._wait(self._idle_wait)
        try:
            data = self._socket.recv(1024)
            if data and len(data):
                self.idle_reset()
                return data
            self.idle()
        except socket.timeout:
            self.idle()
        except socket.error:
            self.idle()
        return None

    def send(self, data):
        self._socket.setblocking(1)
        self._socket.sendall(data)
        self._socket.setblocking(0)

    def close(self):
        message = "Closing Socket - %d" % (self.identify())
        LOG.debug(message)
        try:
            self._socket.shutdown(socket.SHUT_RDWR)
            self._socket.close()
        except Exception as exc:
            message = "%s - exception while closing - %s" % (
                self.identify(), str(exc))
            LOG.error(message)

    def identify(self):
        return self.socket_id


"""
ADT for Proxy Connection Object
Each Connection Object is pair of Unix Socket and
TCP Client Socket
"""


class ProxyConnection(object):

    def __init__(self, conf, unix_socket, tcp_socket):
        self._unix_conn = Connection(conf, unix_socket, type='unix')
        self._tcp_conn = Connection(conf, tcp_socket, type='tcp')
        message = "New Proxy - Unix - %d, TCP - %d" % (
            self._unix_conn.identify(), self._tcp_conn.identify())
        LOG.debug(message)

    def close(self):
        self._unix_conn.close()
        self._tcp_conn.close()

    def _proxy(self, rxconn, txconn):
        data = rxconn.recv()
        if data:
            txconn.send(data)

    def run(self):
        try:
            self._proxy(self._unix_conn, self._tcp_conn)
            self._proxy(self._tcp_conn, self._unix_conn)
            return True
        except Exception as exc:
            message = "%s" % (exc)
            LOG.debug(message)
            self._unix_conn.close()
            self._tcp_conn.close()
            return False

    def identify(self):
        return '%d:%d' % (
            self._unix_conn.identify(),
            self._tcp_conn.identify())


"""
ADT for proxy Worker
"""


class Worker(object):

    def run(self):
        """
        Worker thread will pop the Proxy Connection Object
        from Connection Queue and Perform send and receive
        operations. If the connection is ideal upto ideal_max_timeout
        it will not push the Object into connection queue so Proxy Connection
        Object is automatically destroy, otherwise it will again
        push the Object in connection Queue
        """
        while True:
            try:
                pc = ConnQ.get()
                call = True
                while call:
                    call = pc.run()
            except eventlet.queue.Empty:
                pass
            eventlet.sleep(0)


"""
ADT to  Run the configurator proxy,
        accept the Unix Client request,
        Check REST Server is reachable or not,
        Try to establish TCP Client Connection to REST

"""


class Proxy(object):

    def __init__(self, conf):
        self.conf = conf
        # Be a server and wait for connections from the client
        self.server = UnixServer(conf, self)
        self.client = TcpClient(conf, self)

    def start(self):
        """Run each worker in new thread"""

        for i in range(self.conf.worker_threads):
            eventlet.spawn_n(Worker().run)
        while True:
            self.server.listen()

    def new_client(self, unixsocket, address):
        """Establish connection with the tcp server"""

        tcpsocket, connected = self.client.connect()
        if not connected:
            message = "Proxy -> Could not connect with tcp server"
            LOG.error(message)
            unixsocket.close()
            tcpsocket.close()
        else:
            pc = ProxyConnection(self.conf, unixsocket, tcpsocket)
            ConnQ.put(pc)

PROXY_OPTS = [
    oslo_config.IntOpt(
        'max_connections',
        default=10,
        help='Max num of client connections'
    ),
    oslo_config.IntOpt(
        'worker_threads',
        default=10,
        help='Number of threads to be used.'
    ),
    oslo_config.FloatOpt(
        'connect_max_wait_timeout',
        default=10,
        help='Max time to wait for connect.'
    ),
    oslo_config.FloatOpt(
        'idle_max_wait_timeout',
        default=10,
        help='Max time to keep idle channel open.'
    ),
    oslo_config.FloatOpt(
        'idle_min_wait_timeout',
        default=10,
        help='Minimum time to wait on idle channel.'
    ),
    oslo_config.StrOpt(
        'unix_bind_path',
        default='/var/run/uds_socket',
        help='Path for unix server to bind.'
    ),
    oslo_config.StrOpt(
        'nfp_controller_ip',
        default='172.16.0.3',
        help='NFP controllers ip address'
    ),
    oslo_config.IntOpt(
        'nfp_controller_port',
        default='8070',
        help='NFP controllers port num'
    )
]


def main(argv):
    oslo_config.CONF.register_opts(PROXY_OPTS, 'proxy')
    oslo_config.CONF(args=sys.argv[1:])
    oslo_logging.setup(oslo_config.CONF, 'nfp')
    conf = Configuration(oslo_config.CONF)
    Proxy(conf).start()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg as oslo_config

CONF = oslo_config.CONF

NFP_OPTS = [
    oslo_config.IntOpt(
        'workers',
        default=1,
        help='Number of event worker process to be created.'
    ),
    oslo_config.ListOpt(
        'nfp_modules_path',
        default='gbpservice.nfp.core.test',
        help='Path for NFP modules.'
        'All modules from this path are autoloaded by framework'
    ),
    oslo_config.StrOpt(
        'backend',
        default='rpc',
        help='Backend Support for communicationg with configurator.'
    )
]

EXTRA_OPTS = [
    oslo_config.StrOpt(
        'logger_class',
        default='gbpservice.nfp.core.log.WrappedLogger',
        help='logger class path to handle logging seperately.'
    ),
]


def init(module, args, **kwargs):
    """Initialize the configuration. """
    oslo_config.CONF.register_opts(EXTRA_OPTS)
    oslo_config.CONF.register_opts(NFP_OPTS, module)
    oslo_config.CONF(args=args, project='nfp',
                     version='%%(prog)s %s' % ('version'),
                     **kwargs)

    return oslo_config.CONF



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import copy
import requests
import time

from neutron._i18n import _LI

from neutron._i18n import _LI

from gbpservice.contrib.nfp.configurator.drivers.base import base_driver
from gbpservice.contrib.nfp.configurator.drivers.vpn.vyos import (
    vyos_vpn_constants as const)
from gbpservice.contrib.nfp.configurator.lib import (
    generic_config_constants as gen_cfg_const)
from gbpservice.contrib.nfp.configurator.lib import constants as common_const
from gbpservice.contrib.nfp.configurator.lib import data_parser
from gbpservice.contrib.nfp.configurator.lib import vpn_constants as vpn_const
from gbpservice.nfp.core import log as nfp_logging

from oslo_concurrency import lockutils
from oslo_serialization import jsonutils

LOG = nfp_logging.getLogger(__name__)


class UnknownReasonException(Exception):
    message = "Unsupported rpcreason '%(reason)s' from plugin "


class UnknownResourceException(Exception):
    message = "Unsupported resource '%(resource)s' from plugin "


class RestApi(object):
    """
    Provides different methods to make ReST calls to the service VM,
    to update the configurations
    """

    def __init__(self, vm_mgmt_ip):
        self.vm_mgmt_ip = vm_mgmt_ip
        self.timeout = const.REST_TIMEOUT

    def _dict_to_query_str(self, args):
        return '&'.join([str(k) + '=' + str(v) for k, v in args.iteritems()])

    def post(self, api, args, headers):
        """
        Makes ReST call to the service VM to post the configurations.

        :param api: method that need to called inside the service VM to
        update the configurations.
        :prarm args: data that is need to be configured in service VM

        Returns: None
        """
        url = const.request_url % (
            self.vm_mgmt_ip,
            const.CONFIGURATION_SERVER_PORT, api)
        data = jsonutils.dumps(args)

        try:
            resp = requests.post(url, data=data, timeout=self.timeout,
                                 headers=headers)
            message = jsonutils.loads(resp.text)
            msg = "POST url %s %d" % (url, resp.status_code)
            LOG.debug(msg)
            if resp.status_code == 200 and message.get("status", False):
                msg = "POST Rest API %s - Success" % (url)
                LOG.debug(msg)
            else:
                msg = ("POST Rest API %s - Failed with status %s, %s"
                       % (url, resp.status_code,
                          message.get("reason", None)))
                LOG.error(msg)
                raise requests.exceptions.HTTPError(msg)
        except Exception as err:
            msg = ("Post Rest API %s - Failed. Reason: %s"
                   % (url, str(err).capitalize()))
            LOG.error(msg)
            raise requests.exceptions.HTTPError(msg)

    def put(self, api, args, headers):
        """
        Makes ReST call to the service VM to put the configurations.

        :param api: method that need to called inside the service VM to
        update the configurations.
        :prarm args: data that is need to be configured in service VM

        Returns: None
        """
        url = const.request_url % (
            self.vm_mgmt_ip,
            const.CONFIGURATION_SERVER_PORT, api)
        data = jsonutils.dumps(args)

        try:
            resp = requests.put(url, data=data, timeout=self.timeout,
                                headers=headers)
            msg = "PUT url %s %d" % (url, resp.status_code)
            LOG.debug(msg)
            if resp.status_code == 200:
                msg = "REST API PUT %s succeeded." % url
                LOG.debug(msg)
            else:
                msg = ("REST API PUT %s failed with status: %d."
                       % (url, resp.status_code))
                LOG.error(msg)
        except Exception as err:
            msg = ("REST API for PUT %s failed. %s"
                   % (url, str(err).capitalize()))
            LOG.error(msg)

    def delete(self, api, args, headers, data=None):
        """
        Makes ReST call to the service VM to delete the configurations.

        :param api: method that need to called inside the service VM to
        update the configurations.
        :param args: fixed ip of the service VM to make frame the query string.
        :data args: data that is need to be configured in service VM

        Returns: None
        """
        url = const.request_url % (
            self.vm_mgmt_ip,
            const.CONFIGURATION_SERVER_PORT, api)

        if args:
            url += '?' + self._dict_to_query_str(args)

        if data:
            data = jsonutils.dumps(data)
        try:
            resp = requests.delete(url, timeout=self.timeout, data=data,
                                   headers=headers)
            message = jsonutils.loads(resp.text)
            msg = "DELETE url %s %d" % (url, resp.status_code)
            LOG.debug(msg)
            if resp.status_code == 200 and message.get("status", False):
                msg = "DELETE Rest API %s - Success" % (url)
                LOG.info(msg)
            else:
                msg = ("DELETE Rest API %s - Failed %s"
                       % (url, message.get("reason", None)))
                LOG.error(msg)
                raise requests.exceptions.HTTPError(msg)
        except Exception as err:
            msg = ("Delete Rest API %s - Failed. Reason: %s"
                   % (url, str(err).capitalize()))
            LOG.error(msg)
            raise requests.exceptions.HTTPError(msg)

    def get(self, api, args, headers):
        """
        Makes ReST call to the service VM to put the configurations.

        :param api: method that need to called inside the service VM to
        update the configurations.
        :prarm args: data that is need to be configured in service VM

        Returns: None
        """
        output = ''

        url = const.request_url % (
            self.vm_mgmt_ip,
            const.CONFIGURATION_SERVER_PORT, api)

        try:
            resp = requests.get(url, params=args, timeout=self.timeout,
                                headers=headers)
            msg = "GET url %s %d" % (url, resp.status_code)
            LOG.debug(msg)
            if resp.status_code == 200:
                msg = "REST API GET %s succeeded." % url
                LOG.debug(msg)
                json_resp = resp.json()
                return json_resp
            else:
                msg = ("REST API GET %s failed with status: %d."
                       % (url, resp.status_code))
                LOG.error(msg)
        except requests.exceptions.Timeout as err:
            msg = ("REST API GET %s timed out. %s."
                   % (url, str(err).capitalize()))
            LOG.error(msg)
        except Exception as err:
            msg = ("REST API for GET %s failed. %s"
                   % (url, str(err).capitalize()))
            LOG.error(msg)

        return output


class VPNServiceValidator(object):
    """
    Provides the methods to validate the vpn service which is about to
    be created in order to avoid any conflicts if they exists.
    """

    def __init__(self, agent):
        self.agent = agent

    def _update_service_status(self, vpnsvc, status):
        """
        Driver will call this API to report
        status of VPN service.
        """
        msg = ("Driver informing status: %s."
               % status)
        LOG.debug(msg)
        vpnsvc_status = [{
            'id': vpnsvc['id'],
            'status': status,
            'updated_pending_status':True}]
        return vpnsvc_status

    def _error_state(self, context, vpnsvc, message=''):
        """
        Enqueues the status of the service to ERROR.

        :param context: Dictionary which holds all the required data for
        for vpn service.
        :param vpnsvc: vpn service dictionary.
        :param message: the cause for the error.

        Returns: None
        """
        self.agent.update_status(
            context, self._update_service_status(vpnsvc,
                                                 vpn_const.STATE_ERROR))
        msg = ("Resource vpn service: %r went "
               "to error state, %r" % (vpnsvc['id'], message))
        raise Exception(msg)

    def _active_state(self, context, vpnsvc):
        """
        Enqueues the status of the service to ACTIVE.

        :param context: Dictionary which holds all the required data for
        for vpn service.
        :param vpnsvc: vpn service dictionary.

        Returns: None
        """
        self.agent.update_status(
            context, self._update_service_status(vpnsvc,
                                                 vpn_const.STATE_ACTIVE))

    def _get_local_cidr(self, vpn_svc):
        # REVISIT: position based parsing of description
        svc_desc = vpn_svc['description']
        tokens = svc_desc.split(';')
        local_cidr = tokens[1].split('=')[1]
        return local_cidr

    def validate(self, context, resource_data):
        """
        Get the vpn services for this tenant
        Check for overlapping lcidr - (not allowed)

        :param context: Dictionary which holds all the required data for
        for vpn service.
        :param vpnsvc: vpn service dictionary.

        Returns: None
        """

        vpnsvc = resource_data.get('resource')
        lcidr = resource_data.get('provider_cidr')
        filters = {'tenant_id': [context['tenant_id']]}
        t_vpnsvcs = self.agent.get_vpn_services(
            context, filters=filters)
        vpnsvc.pop("status", None)

        for svc in t_vpnsvcs:
            del svc['status']
        if vpnsvc in t_vpnsvcs:
            t_vpnsvcs.remove(vpnsvc)
        for svc in t_vpnsvcs:
            t_lcidr = self._get_local_cidr(svc)
            if t_lcidr == lcidr:
                msg = ("Local cidr %s conflicts with existing vpn service %s"
                       % (lcidr, svc['id']))
                LOG.error(msg)
                self._error_state(
                    context,
                    vpnsvc, msg)
        self._active_state(context, vpnsvc)


class VpnGenericConfigDriver(base_driver.BaseDriver):
    """
    VPN generic config driver for handling device configurations requests.
    This driver class implements VPN configuration.
    """

    def __init__(self):
        self.timeout = const.REST_TIMEOUT
        self.parse = data_parser.DataParser()

    def _parse_vm_context(self, context):
        try:
            username = str(context['service_vm_context'][
                           'vyos']['username'])
            password = str(context['service_vm_context'][
                           'vyos']['password'])
            headers = {'Content-Type': 'application/json',
                       'username': username,
                       'password': password}
            return headers
        except Exception as e:
            msg = ("Failed to get header from context. ERROR: %s" % e)
            LOG.error(msg)
            raise Exception(msg)

    def configure_healthmonitor(self, context, resource_data):
        vm_status = super(VpnGenericConfigDriver,
                          self).configure_healthmonitor(
                              context, resource_data)
        if resource_data['nfds'][0]['periodicity'] == gen_cfg_const.INITIAL:
            if vm_status == common_const.SUCCESS:
                try:
                    resp = self.configure_user(context, resource_data)
                    if resp not in common_const.SUCCESS_CODES:
                        return common_const.FAILURE
                except Exception as e:
                    msg = ("Failed to configure user. ERROR: %s" % e)
                    LOG.error(msg)
                    return common_const.FAILURE
            return vm_status

    def configure_user(self, context, resource_data):
        headers = self._parse_vm_context(context)
        resource_data = self.parse.parse_data(common_const.HEALTHMONITOR,
                                              resource_data)
        mgmt_ip = resource_data['mgmt_ip']
        url = const.request_url % (mgmt_ip,
                                   self.port,
                                   'change_auth')
        data = {}

        LOG.info(_LI("Initiating POST request to configure Authentication "
                     "service at mgmt ip:%(mgmt_ip)s"),
                 {'mgmt_ip': mgmt_ip})
        err_msg = ("Change Auth POST request to the VyOS firewall "
                   "service at %s failed. " % url)
        try:
            resp = requests.post(url, data=data, headers=headers)
        except Exception as err:
            err_msg += ("Reason: %r" % str(err).capitalize())
            LOG.error(err_msg)
            return err_msg

        if (resp.status_code in common_const.SUCCESS_CODES) and (
            resp.json().get('status') is True):
            msg = ("Configured user authentication successfully"
                   " for vyos service at %r." % mgmt_ip)
            LOG.info(msg)
            return resp.status_code

        err_msg += (("Failed to change Authentication para Status code"
                     ": %r, Reason: %r" %
                     (resp.status_code, resp.json().get('reason')))
                    if type(resp.json()) is dict
                    else ("Reason: " + resp))
        LOG.error(err_msg)
        return err_msg

    def _configure_static_ips(self, context, resource_data):
        """ Configure static IPs for provider and stitching interfaces
        of service VM.

        Issues REST call to service VM for configuration of static IPs.

        :param resource_data: a dictionary of vpn rules and objects
        send by neutron plugin

        Returns: SUCCESS/Failure message with reason.

        """
        headers = self._parse_vm_context(context)
        static_ips_info = dict(
            provider_ip=resource_data.get('provider_ip'),
            provider_cidr=resource_data.get('provider_cidr'),
            provider_mac=resource_data.get('provider_mac'),
            stitching_ip=resource_data.get('stitching_ip'),
            stitching_cidr=resource_data.get('stitching_cidr'),
            stitching_mac=resource_data.get('stitching_mac'))
        mgmt_ip = resource_data['mgmt_ip']

        url = const.request_url % (mgmt_ip,
                                   const.CONFIGURATION_SERVER_PORT,
                                   'add_static_ip')
        data = jsonutils.dumps(static_ips_info)

        msg = ("Initiating POST request to add static IPs for primary "
               "service at: %r" % mgmt_ip)
        LOG.info(msg)
        try:
            resp = requests.post(url, data, timeout=self.timeout,
                                 headers=headers)
        except requests.exceptions.ConnectionError as err:
            msg = ("Failed to establish connection to primary service at: "
                   "%r. ERROR: %r" %
                   (mgmt_ip, str(err).capitalize()))
            LOG.error(msg)
            return msg
        except requests.exceptions.RequestException as err:
            msg = ("Unexpected ERROR happened while adding "
                   "static IPs for primary service at: %r. "
                   "ERROR: %r" %
                   (mgmt_ip, str(err).capitalize()))
            LOG.error(msg)
            return msg

        try:
            result = resp.json()
        except ValueError as err:
            msg = ("Unable to parse response, invalid JSON. URL: "
                   "%r. %r" % (url, str(err).capitalize()))
            LOG.error(msg)
            return msg
        if not result['status']:
            msg = ("Error adding static IPs. URL: %r. Reason: %s." %
                   (url, result['reason']))
            LOG.error(msg)
            return msg

        msg = ("Static IPs successfully added.")
        LOG.info(msg)
        return common_const.STATUS_SUCCESS

    def configure_interfaces(self, context, resource_data):
        """ Configure interfaces for the service VM.

        Calls static IP configuration function and implements
        persistent rule addition in the service VM.
        Issues REST call to service VM for configuration of interfaces.

        :param context: neutron context
        :param resource_data: a dictionary of vpn rules and objects
        send by neutron plugin

        Returns: SUCCESS/Failure message with reason.

        """
        headers = self._parse_vm_context(context)
        resource_data = self.parse.parse_data(common_const.INTERFACES,
                                              resource_data)
        mgmt_ip = resource_data['mgmt_ip']

        try:
            result_log_forward = self._configure_log_forwarding(
                const.request_url, mgmt_ip, self.port, headers)
        except Exception as err:
            msg = ("Failed to configure log forwarding for service at %s. "
                   "Error: %s" % (mgmt_ip, err))
            LOG.error(msg)
        else:
            if result_log_forward == common_const.UNHANDLED:
                pass
            elif result_log_forward != common_const.STATUS_SUCCESS:
                # Failure in log forward configuration won't break chain
                # creation. However, error will be logged for detecting
                # failure.
                msg = ("Failed to configure log forwarding for service at %s."
                       " Error: %s" % (mgmt_ip, result_log_forward))
                LOG.error(msg)

        try:
            result_static_ips = self._configure_static_ips(context,
                                                           resource_data)
        except Exception as err:
            msg = ("Failed to add static IPs. Error: %s" % err)
            LOG.error(msg)
            return msg
        else:
            if result_static_ips != common_const.STATUS_SUCCESS:
                return result_static_ips
            else:
                msg = ("Added static IPs. Result: %s" % result_static_ips)
                LOG.info(msg)

        rule_info = dict(
            provider_mac=resource_data['provider_mac'],
            stitching_mac=resource_data['stitching_mac'])

        url = const.request_url % (mgmt_ip,
                                   const.CONFIGURATION_SERVER_PORT, 'add_rule')
        data = jsonutils.dumps(rule_info)
        msg = ("Initiating POST request to add persistent rule to primary "
               "service at: %r" % mgmt_ip)
        LOG.info(msg)
        try:
            resp = requests.post(url, data, timeout=self.timeout,
                                 headers=headers)
        except requests.exceptions.ConnectionError as err:
            msg = ("Failed to establish connection to primary service at: "
                   "%r. ERROR: %r" %
                   (mgmt_ip, str(err).capitalize()))
            LOG.error(msg)
            return msg
        except requests.exceptions.RequestException as err:
            msg = ("Unexpected ERROR happened  while adding "
                   "persistent rule of primary service at: %r. ERROR: %r" %
                   (mgmt_ip, str(err).capitalize()))
            LOG.error(msg)
            return msg

        try:
            result = resp.json()
        except ValueError as err:
            msg = ("Unable to parse response, invalid JSON. URL: "
                   "%r. %r" % (url, str(err).capitalize()))
            LOG.error(msg)
            return msg
        if not result['status']:
            msg = ("Error adding persistent rule. URL: %r" % url)
            LOG.error(msg)
            return msg

        msg = ("Persistent rule successfully added.")
        LOG.info(msg)
        # wait for 10secs for the ip address to get configured. Sometimes
        # observed that 'set_routes' fail with 'ip not configured' error.
        time.sleep(10)
        return common_const.STATUS_SUCCESS

    def _clear_static_ips(self, context, resource_data):
        """ Clear static IPs for provider and stitching
        interfaces of the service VM.

        Issues REST call to service VM for deletion of static IPs.

        :param resource_data: a dictionary of vpn rules and objects
        send by neutron plugin

        Returns: SUCCESS/Failure message with reason.

        """
        headers = self._parse_vm_context(context)
        static_ips_info = dict(
            provider_ip=resource_data.get('provider_ip'),
            provider_cidr=resource_data.get('provider_cidr'),
            provider_mac=resource_data.get('provider_mac'),
            stitching_ip=resource_data.get('stitching_ip'),
            stitching_cidr=resource_data.get('stitching_cidr'),
            stitching_mac=resource_data.get('stitching_mac'))
        mgmt_ip = resource_data['mgmt_ip']

        url = const.request_url % (mgmt_ip,
                                   const.CONFIGURATION_SERVER_PORT,
                                   'del_static_ip')
        data = jsonutils.dumps(static_ips_info)

        msg = ("Initiating POST request to remove static IPs for primary "
               "service at: %r" % mgmt_ip)
        LOG.info(msg)
        try:
            resp = requests.delete(url, data=data, timeout=self.timeout,
                                   headers=headers)
        except requests.exceptions.ConnectionError as err:
            msg = ("Failed to establish connection to primary service at: "
                   "%r. ERROR: %r" %
                   (mgmt_ip, str(err).capitalize()))
            LOG.error(msg)
            return msg
        except requests.exceptions.RequestException as err:
            msg = ("Unexpected ERROR happened  while removing "
                   "static IPs for primary service at: %r. ERROR: %r" %
                   (mgmt_ip, str(err).capitalize()))
            LOG.error(msg)
            return msg

        try:
            result = resp.json()
        except ValueError as err:
            msg = ("Unable to parse response, invalid JSON. URL: "
                   "%r. %r" % (url, str(err).capitalize()))
            LOG.error(msg)
            return msg
        if not result['status']:
            msg = ("Error removing static IPs. URL: %r. Reason: %s." %
                   (url, result['reason']))
            LOG.error(msg)
            return msg

        msg = ("Static IPs successfully removed.")
        LOG.info(msg)
        return common_const.STATUS_SUCCESS

    def clear_interfaces(self, context, resource_data):
        """ Clear interfaces for the service VM.

        Calls static IP clear function and implements
        persistent rule deletion in the service VM.
        Issues REST call to service VM for deletion of interfaces.

        :param context: neutron context
        :param resource_data: a dictionary of vpn rules and objects
        send by neutron plugin

        Returns: SUCCESS/Failure message with reason.

        """
        headers = self._parse_vm_context(context)
        resource_data = self.parse.parse_data(common_const.INTERFACES,
                                              resource_data)
        try:
            result_static_ips = self._clear_static_ips(context, resource_data)
        except Exception as err:
            msg = ("Failed to remove static IPs. Error: %s" % err)
            LOG.error(msg)
            return msg
        else:
            if result_static_ips != common_const.STATUS_SUCCESS:
                return result_static_ips
            else:
                msg = ("Successfully removed static IPs. "
                       "Result: %s" % result_static_ips)
                LOG.info(msg)

        rule_info = dict(
            provider_mac=resource_data['provider_mac'],
            stitching_mac=resource_data['stitching_mac'])

        mgmt_ip = resource_data['mgmt_ip']

        msg = ("Initiating DELETE persistent rule.")
        LOG.info(msg)
        url = const.request_url % (mgmt_ip,
                                   const.CONFIGURATION_SERVER_PORT,
                                   'delete_rule')

        try:
            data = jsonutils.dumps(rule_info)
            resp = requests.delete(url, data=data, timeout=self.timeout,
                                   headers=headers)
        except requests.exceptions.ConnectionError as err:
            msg = ("Failed to establish connection to service at: %r. "
                   "ERROR: %r" %
                   (mgmt_ip, str(err).capitalize()))
            LOG.error(msg)
            raise Exception(err)
        except requests.exceptions.RequestException as err:
            msg = ("Unexpected ERROR happened  while deleting "
                   "persistent rule of service at: %r. ERROR: %r" %
                   (mgmt_ip, str(err).capitalize()))
            LOG.error(msg)
            raise Exception(err)

        try:
            result = resp.json()
        except ValueError as err:
            msg = ("Unable to parse response, invalid JSON. URL: "
                   "%r. %r" % (url, str(err).capitalize()))
            LOG.error(msg)
            raise Exception(msg)
        if not result['status'] or resp.status_code not in [200, 201, 202]:
            msg = ("Error deleting persistent rule. URL: %r" % url)
            LOG.error(msg)
            raise Exception(msg)
        msg = ("Persistent rule successfully deleted.")
        LOG.info(msg)
        return common_const.STATUS_SUCCESS

    def configure_routes(self, context, resource_data):
        """ Configure routes for the service VM.

        Issues REST call to service VM for configuration of routes.

        :param context: neutron context
        :param resource_data: a dictionary of vpn rules and objects
        send by neutron plugin

        Returns: SUCCESS/Failure message with reason.

        """
        headers = self._parse_vm_context(context)
        forward_routes = resource_data.get('forward_route')
        resource_data = self.parse.parse_data(common_const.ROUTES,
                                              resource_data)
        mgmt_ip = resource_data.get('mgmt_ip')
        gateway_ip = resource_data.get('stitching_gw_ip')

        # checking whether VPN service is present in the chain
        # if yes, just configure the stitching pbr else
        # configure both stitching and provider pbrs.

        if not forward_routes:
            source_cidrs = [resource_data.get('stitching_cidr')]
        else:
            source_cidrs = [resource_data.get('provider_cidr'),
                            resource_data.get('stitching_cidr')]

        stitching_url = const.request_url % (mgmt_ip,
                                             const.CONFIGURATION_SERVER_PORT,
                                             'add-stitching-route')
        st_data = jsonutils.dumps({'gateway_ip': gateway_ip})

        try:
            resp = requests.post(
                stitching_url, data=st_data, timeout=self.timeout,
                headers=headers)
        except requests.exceptions.ConnectionError as err:
            msg = ("Failed to establish connection to service at: "
                   "%r. ERROR: %r" % (mgmt_ip,
                                      str(err).capitalize()))
            LOG.error(msg)
            return msg
        except requests.exceptions.RequestException as err:
            msg = ("Unexpected ERROR happened  while configuring "
                   "default gw route of service at: %r ERROR: %r" %
                   (mgmt_ip, str(err).capitalize()))
            LOG.error(msg)
            return msg

        url = const.request_url % (mgmt_ip, const.CONFIGURATION_SERVER_PORT,
                                   'add-source-route')
        active_configured = False
        route_info = []
        for source_cidr in source_cidrs:
            route_info.append({'source_cidr': source_cidr,
                               'gateway_ip': gateway_ip})
        data = jsonutils.dumps(route_info)
        msg = ("Initiating POST request to configure route of "
               "primary service at: %r" % mgmt_ip)
        LOG.info(msg)
        try:
            resp = requests.post(url, data=data, timeout=self.timeout,
                                 headers=headers)
        except requests.exceptions.ConnectionError as err:
            msg = ("Failed to establish connection to service at: "
                   "%r. ERROR: %r" % (mgmt_ip, str(err).capitalize()))
            LOG.error(msg)
            return msg
        except requests.exceptions.RequestException as err:
            msg = ("Unexpected ERROR happened  while configuring "
                   "route of service at: %r ERROR: %r" %
                   (mgmt_ip, str(err).capitalize()))
            LOG.error(msg)
            return msg

        if resp.status_code in common_const.SUCCESS_CODES:
            message = jsonutils.loads(resp.text)
            if message.get("status", False):
                msg = ("Route configured successfully for VYOS"
                       " service at: %r" % mgmt_ip)
                LOG.info(msg)
                active_configured = True
            else:
                msg = ("Configure source route failed on service with"
                       " status %s %s"
                       % (resp.status_code, message.get("reason", None)))
                LOG.error(msg)
                return msg

        msg = ("Route configuration status : %r "
               % (active_configured))
        LOG.info(msg)
        if active_configured:
            return common_const.STATUS_SUCCESS
        else:
            return ("Failed to configure source route. Response code: %s."
                    "Response Content: %r" % (resp.status_code, resp.content))

    def clear_routes(self, context, resource_data):
        """ Clear routes for the service VM.

        Issues REST call to service VM for deletion of routes.

        :param context: neutron context
        :param resource_data: a dictionary of vpn rules and objects
        send by neutron plugin

        Returns: SUCCESS/Failure message with reason.

        """
        # clear the static stitching gateway route
        headers = self._parse_vm_context(context)
        resource_data = self.parse.parse_data(common_const.ROUTES,
                                              resource_data)
        mgmt_ip = resource_data.get('mgmt_ip')
        source_cidrs = [resource_data.get('provider_cidr'),
                        resource_data.get('stitching_cidr')]

        stitching_url = const.request_url % (mgmt_ip,
                                             const.CONFIGURATION_SERVER_PORT,
                                             'delete-stitching-route')
        st_data = jsonutils.dumps(
            {'gateway_ip': resource_data.get('stitching_gw_ip')})
        try:
            resp = requests.post(
                stitching_url, data=st_data, timeout=self.timeout,
                headers=headers)
        except requests.exceptions.ConnectionError as err:
            msg = ("Failed to establish connection to service at: "
                   "%r. ERROR: %r" % (mgmt_ip,
                                      str(err).capitalize()))
            LOG.error(msg)
            return msg

        active_configured = False
        url = const.request_url % (mgmt_ip, const.CONFIGURATION_SERVER_PORT,
                                   'delete-source-route')
        route_info = []
        for source_cidr in source_cidrs:
            route_info.append({'source_cidr': source_cidr})
        data = jsonutils.dumps(route_info)
        msg = ("Initiating DELETE route request to primary service at: %r"
               % mgmt_ip)
        LOG.info(msg)
        try:
            resp = requests.delete(url, data=data, timeout=self.timeout,
                                   headers=headers)
        except requests.exceptions.ConnectionError as err:
            msg = ("Failed to establish connection to primary service at: "
                   " %r. ERROR: %r" % (mgmt_ip, err))
            LOG.error(msg)
            return msg
        except requests.exceptions.RequestException as err:
            msg = ("Unexpected ERROR happened  while deleting "
                   " route of service at: %r ERROR: %r"
                   % (mgmt_ip, err))
            LOG.error(msg)
            return msg

        if resp.status_code in common_const.SUCCESS_CODES:
            active_configured = True

        msg = ("Route deletion status : %r "
               % (active_configured))
        LOG.info(msg)
        if active_configured:
            return common_const.STATUS_SUCCESS
        else:
            return ("Failed to delete source route. Response code: %s."
                    "Response Content: %r" % (resp.status_code, resp.content))


@base_driver.set_class_attr(SERVICE_TYPE=vpn_const.SERVICE_TYPE,
                            SERVICE_VENDOR=const.SERVICE_VENDOR)
class VpnaasIpsecDriver(VpnGenericConfigDriver):
    """
    Driver class for implementing VPN IPSEC configuration
    requests from VPNaas Plugin.
    """

    def __init__(self, conf):
        self.conf = conf
        self.port = const.CONFIGURATION_SERVER_PORT
        self.handlers = {
            'vpn_service': {
                'create': self.create_vpn_service},
            'ipsec_site_connection': {
                'create': self.create_ipsec_conn,
                'update': self.update_ipsec_conn,
                'delete': self.delete_ipsec_conn}}
        super(VpnaasIpsecDriver, self).__init__()

    def _update_conn_status(self, conn, status):
        """
        Driver will call this API to report
        status of a connection - only if there is any change.
        :param conn: ipsec conn dicitonary
        :param status: status of the service.

        Returns: updated status dictionary
        """
        msg = ("Driver informing connection status "
               "changed to %s" % status)
        LOG.debug(msg)
        vpnsvc_status = [{
            'id': conn['vpnservice_id'],
            'status':'ACTIVE',
            'updated_pending_status':False,
            'ipsec_site_connections':{
                conn['id']: {
                    'status': status,
                    'updated_pending_status': True}}}]
        return vpnsvc_status

    def _error_state(self, context, conn, message=''):
        """
        Enqueues the status of the service to ERROR.

        :param context: Dictionary which holds all the required data for
        for vpn service.
        :param conn: ipsec conn dicitonary.
        :param message: the cause for the error.

        Returns: None
        """

        self.agent.update_status(
            context, self._update_conn_status(conn,
                                              vpn_const.STATE_ERROR))
        msg = ("Resource ipsec site connection: %r went "
               "to error state, %r" % (conn['id'], message))
        raise Exception(msg)

    def _init_state(self, context, conn):
        """
        Enqueues the status of the service to ACTVIE.

        :param context: Dictionary which holds all the required data for
        for vpn service.
        :param conn: ipsec conn dicitonary.

        Returns: None
        """
        msg = "IPSec: Configured successfully- %s " % conn['id']
        LOG.info(msg)
        self.agent.update_status(
            context, self._update_conn_status(conn,
                                              vpn_const.STATE_INIT))

        for item in context['service_info']['ipsec_site_conns']:
            if item['id'] == conn['id']:
                item['status'] = vpn_const.STATE_INIT

    def _ipsec_conn_correct_enc_algo(self, conn):
        ike_enc_algo = conn['ikepolicy']['encryption_algorithm']
        ipsec_enc_algo = conn['ipsecpolicy']['encryption_algorithm']

        algos = {
            'aes-128': "aes128",
            'aes-256': "aes256",
            'aes-192': "aes256"}

        if ike_enc_algo in algos.keys():
            ike_enc_algo = algos[ike_enc_algo]
        if ipsec_enc_algo in algos.keys():
            ipsec_enc_algo = algos[ipsec_enc_algo]

        conn['ikepolicy']['encryption_algorithm'] = ike_enc_algo
        conn['ipsecpolicy']['encryption_algorithm'] = ipsec_enc_algo
        conn['ikepolicy']['name'] = (
            "ike-%s" % conn['ikepolicy']['id'].split('-')[0])
        conn['ipsecpolicy']['name'] = (
            "ipsec-%s" % conn['ikepolicy']['id'].split('-')[0])

    def _get_filters(self, tenant_id=None, vpnservice_id=None, conn_id=None,
                     peer_address=None):
        filters = {}
        if tenant_id:
            filters['tenant_id'] = tenant_id
        if vpnservice_id:
            filters['vpnservice_id'] = vpnservice_id
        if conn_id:
            filters['siteconn_id'] = conn_id
        if peer_address:
            filters['peer_address'] = peer_address
        return filters

    def _get_stitching_cidr(self, conn):
        # REVISIT: position based parsing of description
        desc = conn['description']
        tokens = desc.split(';')
        stitching_cidr = tokens[5].split('=')[1]
        return stitching_cidr

    def _get_access_ip(self, conn):
        svc_desc = conn['description']
        tokens = svc_desc.split(';')
        access_ip = tokens[2].split('=')[1]
        return access_ip

    def _get_ipsec_tunnel_local_cidr_from_vpnsvc(self, vpn_svc):
        svc_desc = vpn_svc['description']
        tokens = svc_desc.split(';')
        tunnel_local_cidr = tokens[1].split('=')[1]

        standby_fip = None
        try:
            standby_fip = tokens[9].split('=')[1]
        except Exception:
            pass
        return tunnel_local_cidr, standby_fip

    def _get_ipsec_tunnel_local_cidr(self, svc_context):
        return self._get_ipsec_tunnel_local_cidr_from_vpnsvc(
            svc_context['service'])

    def _ipsec_create_conn(self, context, mgmt_fip, resource_data):
        """
        Get the context for this ipsec conn and make POST to the service VM.
        :param context: Dictionary which holds all the required data for
        for vpn service.
        :param mgmt_fip: managent floting ip
        :paraM conn: ipsec conn dictionary

        Returns: None
        """
        headers = self._parse_vm_context(context['agent_info']['context'])
        conn = resource_data.get('resource')
        description = conn['description']
        svc_context = self.agent.get_vpn_servicecontext(
            context, self._get_filters(conn_id=conn['id']))[0]
        dhgroup = {'group2': 2,
                   'group5': 5,
                   'group14': 14}
        svc_context['siteconns'][0]['ikepolicy']['pfs_group'] = (
            dhgroup[svc_context['siteconns'][0]['ikepolicy']['pfs']])
        svc_context['siteconns'][0]['ipsecpolicy']['pfs_group'] = (
            dhgroup[svc_context['siteconns'][0]['ipsecpolicy']['pfs']])

        # For cluster we need to send standby_fip to svc vm agent
        tunnel_local_cidr, standby_fip = (
            self._get_ipsec_tunnel_local_cidr_from_vpnsvc(conn))
        if standby_fip:
            svc_context['siteconns'][0]['connection']['standby_fip'] = (
                standby_fip)
        conn = svc_context['siteconns'][0]['connection']
        conn['description'] = description
        svc_context['siteconns'][0]['connection']['stitching_fixed_ip'] = (
            resource_data['stitching_ip'])
        svc_context['siteconns'][0]['connection']['access_ip'] = (
            resource_data['stitching_floating_ip'])
        msg = "IPSec: Pushing ipsec configuration %s" % conn
        LOG.info(msg)
        conn['tunnel_local_cidr'] = tunnel_local_cidr
        self._ipsec_conn_correct_enc_algo(svc_context['siteconns'][0])
        peer_cidrs_from_2 = conn['peer_cidrs'][1:]
        conn['peer_cidrs'] = [conn['peer_cidrs'][0]]
        svc_context['service']['cidr'] = self._get_stitching_cidr(conn)
        RestApi(mgmt_fip).post(
            "create-ipsec-site-conn", svc_context, headers)
        if peer_cidrs_from_2:
            tunnel = {}
            tunnel['peer_address'] = conn['peer_address']
            tunnel['local_cidr'] = tunnel_local_cidr
            tunnel['peer_cidrs'] = peer_cidrs_from_2
            RestApi(mgmt_fip).post(
                "create-ipsec-site-tunnel", tunnel, headers)
        self._init_state(context, conn)

    def _ipsec_create_tunnel(self, context, mgmt_fip, conn):
        """
        Get the context for this ipsec conn and make POST to the service VM.
        :param context: Dictionary which holds all the required data for
        for vpn service.
        :param mgmt_fip: managent floting ip
        :paraM conn: ipsec conn dictionary

        Returns: None
        """
        headers = self._parse_vm_context(context['agent_info']['context'])
        tunnel_local_cidr, _ = (
            self._get_ipsec_tunnel_local_cidr_from_vpnsvc(conn))

        tunnel = {}
        tunnel['peer_address'] = conn['peer_address']
        tunnel['local_cidr'] = tunnel_local_cidr
        tunnel['peer_cidrs'] = conn['peer_cidrs']
        RestApi(mgmt_fip).post(
            "create-ipsec-site-tunnel", tunnel, headers)
        self._init_state(context, conn)

    def _ipsec_get_tenant_conns(self, context, mgmt_fip, conn,
                                on_delete=False):
        """
        Get the context for this ipsec conn and vpn services.

        :param context: Dictionary which holds all the required data for
        for vpn service.
        :param mgmt_fip: managent floting ip
        :paraM conn: ipsec conn dictionary

        Returns: list of ipsec conns
        """

        filters = {
            'tenant_id': [context['tenant_id']],
            'peer_address': [conn['peer_address']]}
        tenant_conns = self.agent.get_ipsec_conns(
            context, filters)
        if not tenant_conns:
            if not on_delete:
                # Something went wrong - atleast the current
                # connection should be there
                msg = "No tenant conns for filters (%s)" % (str(filters))
                LOG.error(msg)
                # Move conn into err state
                self._error_state(context, conn, msg)

        conn_to_remove = None

        for connection in tenant_conns:
            if connection['id'] == conn['id']:
                conn_to_remove = connection
                break
        if conn_to_remove:
            tenant_conns.remove(conn_to_remove)
        if not tenant_conns:
            return tenant_conns

        conn_list = []
        # get fip from connn description
        access_ip = self._get_access_ip(conn)
        svc_ids = [conn['vpnservice_id'] for conn in tenant_conns]
        vpnservices = self.agent.get_vpn_services(context, ids=svc_ids)
        copy_svc = copy.deepcopy(vpnservices)
        # if service's fip matches new service's fip then both services
        # lie on same instance, in this case we should only create tunnel
        for vpn in copy_svc:
            if access_ip in vpn['description']:
                continue
            else:
                vpnservices.remove(vpn)
        # we have all the vpnservices matching on this fip with same peer
        for vpn in vpnservices:
            # check any connection with same local(provider) subnet
            matching_conn = [conn for conn in tenant_conns
                             if conn['vpnservice_id'] == vpn['id']]
            conn_list.extend(matching_conn)
        if not on_delete:
            # Remove the conns which are in pending_create
            # state. It might be possible that more than one
            # conns could get created in database before the rpc
            # method of dev driver is invoked.
            # We have to separate first conn creation from rest.
            copy_conns = copy.deepcopy(conn_list)
            for tconn in copy_conns:
                if tconn['status'] == (
                        vpn_const.STATE_PENDING and tconn in conn_list):
                    conn_list.remove(tconn)
        # conn_list is list of site connections which share same vpn_service.
        return conn_list

    def _ipsec_check_overlapping_peer(self, context,
                                      tenant_conns, conn):
        pcidrs = conn['peer_cidrs']
        peer_address = conn['peer_address']
        for t_conn in tenant_conns:
            t_pcidrs = t_conn['peer_cidrs']
            if conn['vpnservice_id'] != t_conn['vpnservice_id']:
                continue

            for pcidr in pcidrs:
                if (pcidr in t_pcidrs) and (
                        not peer_address == t_conn['peer_address']):
                    msg = "Overlapping peer cidr (%s)" % (pcidr)
                    LOG.error(msg)
                    self._error_state(
                        context, conn, msg)

    def _ipsec_delete_tunnel(self, context, mgmt_fip,
                             resource_data):
        """
        Make DELETE to the service VM to delete the tunnel.

        :param context: Dictionary which holds all the required data for
        for vpn service.
        :param mgmt_fip: managent floting ip
        :paraM conn: ipsec conn dictionary

        Returns: None
        """
        headers = self._parse_vm_context(context['agent_info']['context'])
        conn = resource_data.get('resource')
        lcidr = resource_data['provider_cidr']

        tunnel = {}
        tunnel['peer_address'] = conn['peer_address']
        tunnel['local_cidr'] = lcidr
        tunnel['peer_cidrs'] = conn['peer_cidrs']
        try:
            RestApi(mgmt_fip).delete(
                "delete-ipsec-site-tunnel", tunnel, headers)
            self.agent.ipsec_site_conn_deleted(context, conn['id'])
        except Exception as err:
            msg = ("IPSec: Failed to delete IPSEC tunnel. %s"
                   % str(err).capitalize())
            LOG.error(msg)

    def _ipsec_delete_connection(self, context, mgmt_fip,
                                 conn):
        """
        Make DELETE to the service VM to delete the ipsec conn.

        :param context: Dictionary which holds all the required data for
        for vpn service.
        :param mgmt_fip: managent floting ip
        :paraM conn: ipsec conn dictionary

        Returns: None
        """

        try:
            headers = self._parse_vm_context(context['agent_info']['context'])
            RestApi(mgmt_fip).delete(
                "delete-ipsec-site-conn",
                {'peer_address': conn['peer_address']}, headers)
            self.agent.ipsec_site_conn_deleted(context, conn['id'])
        except Exception as err:
            msg = ("IPSec: Failed to delete IPSEC conn. %s"
                   % str(err).capitalize())
            LOG.error(msg)

    def _ipsec_is_state_changed(self, context, svc_context, conn, fip):
        """
        Make GET request to the service VM to get the status of the site conn.

        :param svc_context: list of ipsec conn dictionaries
        :paraM conn: ipsec conn dictionary
        :param fip: floting ip of the service VM

        Returns: None
        """
        headers = self._parse_vm_context(context['agent_info']['context'])
        c_state = None
        lcidr, _ = self._get_ipsec_tunnel_local_cidr_from_vpnsvc(conn)
        if conn['status'] == vpn_const.STATE_INIT:
            tunnel = {
                'peer_address': conn['peer_address'],
                'local_cidr': lcidr,
                'peer_cidr': conn['peer_cidrs'][0]}
            output = RestApi(fip).get(
                "get-ipsec-site-tunnel-state",
                tunnel, headers)
            state = output['state']

            if state.upper() == 'UP' and (
               conn['status'] != vpn_const.STATE_ACTIVE):
                c_state = vpn_const.STATE_ACTIVE
            if state.upper() == 'DOWN' and (
               conn['status'] == vpn_const.STATE_ACTIVE):
                c_state = vpn_const.STATE_PENDING

        if c_state:
            return c_state, True
        return c_state, False

    def create_vpn_service(self, context, resource_data):
        msg = "Validating VPN service %s " % resource_data.get('resource')
        LOG.info(msg)
        validator = VPNServiceValidator(self.agent)
        validator.validate(context, resource_data)

    def create_ipsec_conn(self, context, resource_data):
        """
        Implements functions to make update ipsec configuration in service VM.

        :param context: context dictionary of vpn service type
        :param resource_data: dicionary of a specific operation type,
             which was sent from neutron plugin

        Returns: None
        """
        conn = resource_data.get('resource')
        mgmt_fip = resource_data['mgmt_ip']
        msg = "IPsec: create site connection %s" % conn
        LOG.info(msg)
        """
        Following conditions -
        0) Conn with more than one peer_address
        is not allowed. This is because vyos has
        conns and tunnels inside conn. But openstack
        doesnt have tunnels. So conn will itslef need
        to be mapped to tunnel.
        a) Already conns exist for this tenant
            . In this case just add a tunnel
                . For same peer
                . Add peer for different peer
        b) First conn, create complete ipsec profile
        """
        t_lcidr = resource_data['provider_cidr']
        if t_lcidr in conn['peer_cidrs']:
            msg = ("IPSec: Tunnel remote cidr %s conflicts "
                   "with local cidr." % t_lcidr)
            LOG.error(msg)
            self._error_state(context, conn, msg)
        if len(conn['peer_cidrs']) < 1:
            msg = ("IPSec: Invalid number of peer CIDR. Should not be"
                   " less than 1.")
            LOG.error(msg)
            self._error_state(context, conn, msg)

        try:
            tenant_conns = self._ipsec_get_tenant_conns(
                context, mgmt_fip, conn)
        except Exception as err:
            msg = ("IPSec: Failed to get tenant conns for IPSEC create. %s"
                   % str(err).capitalize())
            LOG.error(msg)
            self._error_state(context, conn, msg)
        try:
            """
            Check if this conn has overlapping peer
            cidr with any other conn for the same
            tenant - we do not support this model.
            """
            self._ipsec_check_overlapping_peer(
                context, tenant_conns, conn)
            self._ipsec_create_conn(context, mgmt_fip, resource_data)

        except Exception as ex:
            msg = "IPSec: Exception in creating ipsec conn: %s" % ex
            LOG.error(msg)
            self._error_state(context, conn, msg)

    def update_ipsec_conn(self, context, resource_data):
        """
        Implements functions to make update ipsec configuration in service VM.

        :param context: context dictionary of vpn service type
        :param resource_data: dicionary of a specific operation type,
             which was sent from neutron plugin

        Returns: None
        """
        pass

    def delete_ipsec_conn(self, context, resource_data):
        """
        Implements function to make delete ipsec configuration in service VM.

        :param context: context dictionary of vpn service type
        :param resource_data: dicionary of a specific operation type,
             which was sent from neutron plugin

        Returns: None
        """
        conn = resource_data.get('resource')
        msg = "IPsec: delete siteconnection %s" % conn
        LOG.info(msg)
        mgmt_fip = resource_data['mgmt_ip']

        tenant_conns = self._ipsec_get_tenant_conns(
            context, mgmt_fip, conn, on_delete=True)
        try:
            if tenant_conns:
                self._ipsec_delete_tunnel(
                    context, mgmt_fip, resource_data)
            else:
                self._ipsec_delete_connection(
                    context, mgmt_fip, conn)
        except Exception as ex:
            msg = "IPSec: delete ipsec conn failed %s " % ex
            LOG.error(msg)
            self._error_state(context, conn, msg)

    def check_status(self, context, svc_context):
        """
        Implements functions to get the status of the site to site conn.

        :param context: context dictionary of vpn service type
        :param svc_contex: list of ipsec conn dictionaries

        Returns: None
        """
        vpn_desc = self.parse.parse_data(common_const.VPN, context)
        # Other than non HA vpn_desc will be a list of parsed nfs
        if type(vpn_desc) == list:
            fip = vpn_desc[0]['mgmt_ip']
        else:
            fip = vpn_desc['mgmt_ip']

        conn = svc_context['siteconns'][0]['connection']

        try:
            state, changed = self._ipsec_is_state_changed(context,
                svc_context, conn, fip)
        except Exception as err:
            msg = ("Failed to check if IPSEC state is changed. %s"
                   % str(err).capitalize())
            LOG.error(msg)
            return vpn_const.STATE_ERROR
        if changed:
            self.agent.update_status(
                context, self._update_conn_status(conn,
                                                  state))
        return state

    def vpnservice_updated(self, context, resource_data):
        """
        Demultiplexes the different methods to update the configurations

        :param context: context dictionary of vpn service type
        :param resource_data: dicionary of a specific operation type,
             which was sent from neutron plugin

        Returns: None
        """
        vpn_desc = self.parse.parse_data(common_const.VPN, context)
        resource_data.update(vpn_desc)

        msg = ("Handling VPN service update notification for '%s'"
               % resource_data.get('reason', ''))
        LOG.info(msg)

        resource = resource_data.get('resource')
        tenant_id = resource['tenant_id']
        # Synchronize the update operation per tenant.
        # Resources under tenant have inter dependencies.

        @lockutils.synchronized(tenant_id)
        def _vpnservice_updated(context, resource_data):
            reason = resource_data.get('reason')
            rsrc = resource_data.get('rsrc_type')

            if rsrc not in self.handlers.keys():
                raise UnknownResourceException(rsrc=rsrc)

            if reason not in self.handlers[rsrc].keys():
                raise UnknownReasonException(reason=reason)

            self.handlers[rsrc][reason](context, resource_data)

        return _vpnservice_updated(context, resource_data)



# Copyright (c) 2013 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.contrib.nfp.config_orchestrator.common import topics
from gbpservice.contrib.nfp.configurator.drivers.loadbalancer.v1.haproxy\
    import haproxy_lb_driver
from neutron.api.v2 import attributes
from neutron.plugins.common import constants
from neutron_lbaas.db.loadbalancer import loadbalancer_db as lb_db
from neutron_lbaas.extensions import loadbalancer
from neutron_lbaas.services.loadbalancer.drivers.common import (
    agent_driver_base as adb
)
from neutron_lib import constants as n_constants
from neutron_lib import exceptions as n_exc
from oslo_db import exception
from oslo_utils import excutils
from oslo_utils import uuidutils


class HaproxyOnVMPluginDriver(adb.AgentDriverBase):
    device_driver = haproxy_lb_driver.DRIVER_NAME

    def __init__(self, plugin):
        # Monkey patch LB agent topic and LB agent type
        adb.l_const.LOADBALANCER_AGENT = topics.LB_NFP_CONFIGAGENT_TOPIC
        adb.q_const.AGENT_TYPE_LOADBALANCER = 'NFP Loadbalancer agent'

        super(HaproxyOnVMPluginDriver, self).__init__(plugin)


def _nfp_create_port_for_vip(self, context, vip_db, subnet_id, ip_address):
        # resolve subnet and create port
        subnet = self._core_plugin.get_subnet(context, subnet_id)
        fixed_ip = {'subnet_id': subnet['id']}
        if ip_address and ip_address != attributes.ATTR_NOT_SPECIFIED:
            fixed_ip['ip_address'] = ip_address
            if subnet.get('gateway_ip') == ip_address:
                raise n_exc.IpAddressInUse(net_id=subnet['network_id'],
                                           ip_address=ip_address)

        port_data = {
            'tenant_id': vip_db.tenant_id,
            'name': 'vip-' + vip_db.id,
            'network_id': subnet['network_id'],
            'mac_address': attributes.ATTR_NOT_SPECIFIED,
            'admin_state_up': False,
            'device_id': '',
            'device_owner': n_constants.DEVICE_OWNER_LOADBALANCER,
            'fixed_ips': [fixed_ip]
        }

        port = self._core_plugin.create_port(context, {'port': port_data})
        vip_db.port_id = port['id']
        with context.session.begin(subtransactions=True):
                vip = self._get_resource(context, lb_db.Vip, vip_db.id)
                vip.update({'port_id': port['id']})
                context.session.flush()

        # explicitly sync session with db
        # context.session.flush()
        vip_db = self._get_resource(context, lb_db.Vip, vip_db.id)

lb_db.LoadBalancerPluginDb._create_port_for_vip = _nfp_create_port_for_vip


def nfp_create_vip(self, context, vip):
    v = vip['vip']
    tenant_id = v['tenant_id']

    with context.session.begin(subtransactions=True):
        if v['pool_id']:
            pool = self._get_resource(context, lb_db.Pool, v['pool_id'])
            # validate that the pool has same tenant
            if pool['tenant_id'] != tenant_id:
                raise n_exc.NotAuthorized()
            # validate that the pool has same protocol
            if pool['protocol'] != v['protocol']:
                raise loadbalancer.ProtocolMismatch(
                    vip_proto=v['protocol'],
                    pool_proto=pool['protocol'])
            if pool['status'] == constants.PENDING_DELETE:
                raise loadbalancer.StateInvalid(state=pool['status'],
                                                id=pool['id'])
        vip_db = lb_db.Vip(id=uuidutils.generate_uuid(),
                           tenant_id=tenant_id,
                           name=v['name'],
                           description=v['description'],
                           port_id=None,
                           protocol_port=v['protocol_port'],
                           protocol=v['protocol'],
                           pool_id=v['pool_id'],
                           connection_limit=v['connection_limit'],
                           admin_state_up=v['admin_state_up'],
                           status=constants.PENDING_CREATE)

        session_info = v['session_persistence']

        if session_info:
            s_p = self._create_session_persistence_db(
                session_info,
                vip_db['id'])
            vip_db.session_persistence = s_p

        try:
            context.session.add(vip_db)
            context.session.flush()
        except exception.DBDuplicateEntry:
            raise loadbalancer.VipExists(pool_id=v['pool_id'])

    try:
        # create a port to reserve address for IPAM
        # do it outside the transaction to avoid rpc calls
        self._create_port_for_vip(
            context, vip_db, v['subnet_id'], v.get('address'))
    except Exception:
        # catch any kind of exceptions
        with excutils.save_and_reraise_exception():
            context.session.delete(vip_db)
            context.session.flush()

    if v['pool_id']:
        # fetching pool again
        pool = self._get_resource(context, lb_db.Pool, v['pool_id'])
        # (NOTE): we rely on the fact that pool didn't change between
        # above block and here
        vip_db['pool_id'] = v['pool_id']
        pool['vip_id'] = vip_db['id']

    vip_db = self._get_resource(context, lb_db.Vip, vip_db['id'])
    return self._make_vip_dict(vip_db)

lb_db.LoadBalancerPluginDb.create_vip = nfp_create_vip



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.contrib.nfp.config_orchestrator.common import topics
from gbpservice.contrib.nfp.configurator.drivers.loadbalancer.v2.\
    haproxy import haproxy_driver_constants
from neutron_lbaas.drivers.common import agent_driver_base as adb


class HaproxyOnVMPluginDriver(adb.AgentDriverBase):
    device_driver = haproxy_driver_constants.DRIVER_NAME

    def __init__(self, plugin):
        # Monkey patch LB agent topic and LB agent type
        adb.lb_const.LOADBALANCER_AGENTV2 = topics.LBV2_NFP_CONFIGAGENT_TOPIC
        adb.lb_const.AGENT_TYPE_LOADBALANCERV2 = 'NFP Loadbalancer V2 agent'

        super(HaproxyOnVMPluginDriver, self).__init__(plugin)



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Admin owns SCI (admin_owns_sci)

Revision ID: c2a9d04c8cef

"""

# revision identifiers, used by Alembic.
revision = 'c2a9d04c8cef'
down_revision = '1fadeb573886'


from alembic import op
import sqlalchemy as sa


def upgrade():

    op.add_column(
        'gpm_ptgs_servicechain_mapping',
        sa.Column('tenant_id', sa.String(length=255), nullable=True)
    )


def downgrade():
    op.drop_column('gpm_ptgs_servicechain_mapping', 'tenant_id')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron._i18n import _LE
from neutron._i18n import _LI
from neutron.common import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log
from oslo_utils import excutils
import stevedore

from gbpservice.neutron.services.grouppolicy.common import exceptions as gp_exc


LOG = log.getLogger(__name__)


class ExtensionManager(stevedore.named.NamedExtensionManager):
    """Manage extension drivers using drivers."""

    def __init__(self):
        # Ordered list of extension drivers, defining
        # the order in which the drivers are called.
        self.ordered_ext_drivers = []

        LOG.info(_LI("Configured extension driver names: %s"),
                 cfg.CONF.group_policy.extension_drivers)
        super(ExtensionManager, self).__init__(
            'gbpservice.neutron.group_policy.extension_drivers',
            cfg.CONF.group_policy.extension_drivers,
            invoke_on_load=True,
            name_order=True)
        LOG.info(_LI("Loaded extension driver names: %s"), self.names())
        self._register_drivers()

    def _register_drivers(self):
        """Register all extension drivers.

        This method should only be called once in the ExtensionManager
        constructor.
        """
        for ext in self:
            self.ordered_ext_drivers.append(ext)
        LOG.info(_LI("Registered extension drivers: %s"),
                 [driver.name for driver in self.ordered_ext_drivers])

    def initialize(self):
        # Initialize each driver in the list.
        for driver in self.ordered_ext_drivers:
            LOG.info(_LI("Initializing extension driver '%s'"), driver.name)
            driver.obj.initialize()

    def extension_aliases(self):
        exts = []
        for driver in self.ordered_ext_drivers:
            alias = driver.obj.extension_alias
            exts.append(alias)
            LOG.info(_LI("Got %(alias)s extension from driver '%(drv)s'"),
                     {'alias': alias, 'drv': driver.name})
        return exts

    def _call_on_ext_drivers(self, method_name, session, data, result):
        """Helper method for calling a method across all extension drivers."""
        for driver in self.ordered_ext_drivers:
            try:
                getattr(driver.obj, method_name)(session, data, result)
            except (gp_exc.GroupPolicyException, n_exc.NeutronException):
                with excutils.save_and_reraise_exception():
                    LOG.exception(
                        _LE("Extension driver '%(name)s' "
                            "failed in %(method)s"),
                        {'name': driver.name, 'method': method_name}
                    )
            except Exception:
                LOG.exception(_LE("Extension driver '%(name)s' "
                                  "failed in %(method)s"),
                              {'name': driver.name, 'method': method_name})
                # We are replacing a non-GBP/non-Neutron exception here
                raise gp_exc.GroupPolicyDriverError(method=method_name)

    def process_create_policy_target(self, session, data, result):
        """Call all extension drivers during PT creation."""
        self._call_on_ext_drivers("process_create_policy_target",
                                  session, data, result)

    def process_update_policy_target(self, session, data, result):
        """Call all extension drivers during PT update."""
        self._call_on_ext_drivers("process_update_policy_target",
                                  session, data, result)

    def extend_policy_target_dict(self, session, result):
        """Call all extension drivers to extend PT dictionary."""
        for driver in self.ordered_ext_drivers:
            driver.obj.extend_policy_target_dict(session, result)

    def process_create_policy_target_group(self, session, data, result):
        """Call all extension drivers during PTG creation."""
        self._call_on_ext_drivers("process_create_policy_target_group",
                                  session, data, result)

    def process_update_policy_target_group(self, session, data, result):
        """Call all extension drivers during PTG update."""
        self._call_on_ext_drivers("process_update_policy_target_group",
                                  session, data, result)

    def extend_policy_target_group_dict(self, session, result):
        """Call all extension drivers to extend PTG dictionary."""
        for driver in self.ordered_ext_drivers:
            driver.obj.extend_policy_target_group_dict(session, result)

    def process_create_application_policy_group(self, session, data, result):
        """Call all extension drivers during PTG creation."""
        self._call_on_ext_drivers("process_create_application_policy_group",
                                  session, data, result)

    def process_update_application_policy_group(self, session, data, result):
        """Call all extension drivers during PTG update."""
        self._call_on_ext_drivers("process_update_application_policy_group",
                                  session, data, result)

    def extend_application_policy_group_dict(self, session, result):
        """Call all extension drivers to extend PTG dictionary."""
        for driver in self.ordered_ext_drivers:
            driver.obj.extend_application_policy_group_dict(session, result)

    def process_create_l2_policy(self, session, data, result):
        """Call all extension drivers during L2P creation."""
        self._call_on_ext_drivers("process_create_l2_policy",
                                  session, data, result)

    def process_update_l2_policy(self, session, data, result):
        """Call all extension drivers during L2P update."""
        self._call_on_ext_drivers("process_update_l2_policy",
                                  session, data, result)

    def extend_l2_policy_dict(self, session, result):
        """Call all extension drivers to extend L2P dictionary."""
        for driver in self.ordered_ext_drivers:
            driver.obj.extend_l2_policy_dict(session, result)

    def process_create_l3_policy(self, session, data, result):
        """Call all extension drivers during L3P creation."""
        self._call_on_ext_drivers("process_create_l3_policy",
                                  session, data, result)

    def process_update_l3_policy(self, session, data, result):
        """Call all extension drivers during L3P update."""
        self._call_on_ext_drivers("process_update_l3_policy",
                                  session, data, result)

    def extend_l3_policy_dict(self, session, result):
        """Call all extension drivers to extend L3P dictionary."""
        for driver in self.ordered_ext_drivers:
            driver.obj.extend_l3_policy_dict(session, result)

    def process_create_policy_classifier(self, session, data, result):
        """Call all extension drivers during PC creation."""
        self._call_on_ext_drivers("process_create_policy_classifier",
                                  session, data, result)

    def process_update_policy_classifier(self, session, data, result):
        """Call all extension drivers during PC update."""
        self._call_on_ext_drivers("process_update_policy_classifier",
                                  session, data, result)

    def extend_policy_classifier_dict(self, session, result):
        """Call all extension drivers to extend PC dictionary."""
        for driver in self.ordered_ext_drivers:
            driver.obj.extend_policy_classifier_dict(session, result)

    def process_create_policy_action(self, session, data, result):
        """Call all extension drivers during PA creation."""
        self._call_on_ext_drivers("process_create_policy_action",
                                  session, data, result)

    def process_update_policy_action(self, session, data, result):
        """Call all extension drivers during PA update."""
        self._call_on_ext_drivers("process_update_policy_action",
                                  session, data, result)

    def extend_policy_action_dict(self, session, result):
        """Call all extension drivers to extend PA dictionary."""
        for driver in self.ordered_ext_drivers:
            driver.obj.extend_policy_action_dict(session, result)

    def process_create_policy_rule(self, session, data, result):
        """Call all extension drivers during PR creation."""
        self._call_on_ext_drivers("process_create_policy_rule",
                                  session, data, result)

    def process_update_policy_rule(self, session, data, result):
        """Call all extension drivers during PR update."""
        self._call_on_ext_drivers("process_update_policy_rule",
                                  session, data, result)

    def extend_policy_rule_dict(self, session, result):
        """Call all extension drivers to extend PR dictionary."""
        for driver in self.ordered_ext_drivers:
            driver.obj.extend_policy_rule_dict(session, result)

    def process_create_policy_rule_set(self, session, data, result):
        """Call all extension drivers during PRS creation."""
        self._call_on_ext_drivers("process_create_policy_rule_set",
                                  session, data, result)

    def process_update_policy_rule_set(self, session, data, result):
        """Call all extension drivers during PRS update."""
        self._call_on_ext_drivers("process_update_policy_rule_set",
                                  session, data, result)

    def extend_policy_rule_set_dict(self, session, result):
        """Call all extension drivers to extend PRS dictionary."""
        for driver in self.ordered_ext_drivers:
            driver.obj.extend_policy_rule_set_dict(session, result)

    def process_create_network_service_policy(self, session, data, result):
        """Call all extension drivers during NSP creation."""
        self._call_on_ext_drivers("process_create_network_service_policy",
                                  session, data, result)

    def process_update_network_service_policy(self, session, data, result):
        """Call all extension drivers during NSP update."""
        self._call_on_ext_drivers("process_update_network_service_policy",
                                  session, data, result)

    def extend_network_service_policy_dict(self, session, result):
        """Call all extension drivers to extend NSP dictionary."""
        for driver in self.ordered_ext_drivers:
            driver.obj.extend_network_service_policy_dict(session, result)

    def process_create_external_segment(self, session, data, result):
        """Call all extension drivers during EP creation."""
        self._call_on_ext_drivers("process_create_external_segment",
                                  session, data, result)

    def process_update_external_segment(self, session, data, result):
        """Call all extension drivers during EP update."""
        self._call_on_ext_drivers("process_update_external_segment",
                                  session, data, result)

    def extend_external_segment_dict(self, session, result):
        """Call all extension drivers to extend EP dictionary."""
        for driver in self.ordered_ext_drivers:
            driver.obj.extend_external_segment_dict(session, result)

    def process_create_external_policy(self, session, data, result):
        """Call all extension drivers during EP creation."""
        self._call_on_ext_drivers("process_create_external_policy",
                                  session, data, result)

    def process_update_external_policy(self, session, data, result):
        """Call all extension drivers during EP update."""
        self._call_on_ext_drivers("process_update_external_policy",
                                  session, data, result)

    def extend_external_policy_dict(self, session, result):
        """Call all extension drivers to extend EP dictionary."""
        for driver in self.ordered_ext_drivers:
            driver.obj.extend_external_policy_dict(session, result)

    def process_create_nat_pool(self, session, data, result):
        """Call all extension drivers during NP creation."""
        self._call_on_ext_drivers("process_create_nat_pool",
                                  session, data, result)

    def process_update_nat_pool(self, session, data, result):
        """Call all extension drivers during NP update."""
        self._call_on_ext_drivers("process_update_nat_pool",
                                  session, data, result)

    def extend_nat_pool_dict(self, session, result):
        """Call all extension drivers to extend NP dictionary."""
        for driver in self.ordered_ext_drivers:
            driver.obj.extend_nat_pool_dict(session, result)
